"""Tools to monitor training/validation processes."""

import glob
from functools import partial

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from plotly import colors as pcolors
from plotly import graph_objs as go
from plotly import subplots as psubplots
from plotly.offline import iplot

from .layout import PLOTLY_COLORS_TUPLE, apply_latex_style, color_rgba

__all__ = ["TrainDrawer"]


class TrainDrawer:
    """Class which centralizes function used to monitor a training process."""

    def __init__(
        self,
        log_dir,
        interactive=True,
        paper=False,
        alpha=0.5,
        train_prefix="train",
        val_prefix="inference",
        separator=":",
    ):
        """Initialize the drawer.

        Parameters
        ----------
        log_dir, str
            Path to the parent directory of all the log files
        interactive : bool, default True
            If `True`, use plotly to draw the training/validation curve
        paper : bool, default False
            If `True`, format the figure for a paper, using latext style
        alpha : float, default 0.5
            Opacity of the traces
        train_prefix : str, default 'train'
            Log name prefix shared between training logs
        val_prefix : str, default 'inference'
            Log name prefix shared between validation logs
        separator : str, default ':'
            Character used to separate the acceptable metric names in the
            metric parameter
        """
        # Store the path to the main log directory
        self.log_dir = log_dir
        self.train_prefix = train_prefix
        self.val_prefix = val_prefix
        self.separator = separator

        # Initialize the style
        self.interactive = interactive
        if interactive:
            self.layout = None
            self.cr_char = "<br>"
            self.initialize_plotly()
        else:
            self.linwidth = None
            self.markersize = None
            self.cr_char = "\n"
            self.initialize_matplotlib(paper)

        # Initialize the colors
        self.alpha = alpha
        if interactive:
            self.colors = pcolors.color_parser(
                PLOTLY_COLORS_TUPLE, partial(color_rgba, alpha=alpha)
            )
        else:
            self.colors = PLOTLY_COLORS_TUPLE

    def initialize_plotly(self):
        """Initialize the style parameters for plotly."""
        font = {"size": 20}
        axis_base = {"tickfont": font, "linecolor": "black", "mirror": True}
        self.layout = go.Layout(
            template="plotly_white",
            width=1000,
            height=500,
            margin={"t": 20, "b": 20, "l": 20, "r": 20},
            xaxis={"title": {"text": "Epochs", "font": font}, **axis_base},
            yaxis={"title": {"text": "Metric", "font": font}, **axis_base},
            legend={"font": font, "tracegroupgap": 1},
        )

    def initialize_matplotlib(self, paper):
        """Initialize the style parameters for matplotlib.

        paper : bool, default False
            If `True`, format the figure for a paper, using latext style
        """
        # Define a matplotlib layout using seaborn
        if paper:
            # Define the style for latex
            apply_latex_style()
            self.linewidth = 0.5
            self.markersize = 1

        else:
            # Define the style to be displayed in a notebook
            sns.set(rc={"figure.figsize": (9, 6)}, context="notebook", font_scale=2)
            sns.set_style("white")
            sns.set_style(rc={"axes.grid": True})
            self.linewidth = 2
            self.markersize = 10

    def set_log_dir(self, log_dir):
        """Simply reset the base log directory to another one.

        Parameters
        ----------
        log_dir, str
            Path to the parent directory of all the log files
        """
        self.log_dir = log_dir

    def draw(
        self,
        model,
        metric,
        limits=None,
        model_name=None,
        metric_name=None,
        max_iter=None,
        step=None,
        smoothing=None,
        iter_per_epoch=None,
        print_min=False,
        print_max=False,
        same_plot=True,
        leg_ncols=1,
        figure_name=None,
    ):
        """Finds all training and validation log files inside the specified
        directory and draws an evolution plot of the requested quantities.

        Parameters
        ----------
        model : Union[str, List[str]]
            Model (folder) / list of models names under the main directory
        metric : Union[str, List[str]]
            Metric / list of metrics to draw
        limits : Union[List[float], Dict[str, List[float]]], optional
            List of y boundaries for the plot. If specified as a dictionary,
            can be used to specify different boundaries for each metric.
        model_name : Union[str, Dict[str, str]], optional
            Name of the model as displayed in the legend. If there are multiple
            models, provide a dictionary which maps each model onto a name
        metric_name : Union[str, Dict[str, str]], optional
            Name of the metric as displayed in the legend. If there are multiple
            metrics, provide a dictionary which maps each metric onto a name
        max_iter : int, optional
            Maximum number of iterations to include in the plot
        step : int, optional
            Step between two consecutive iterations that are represented
        smoothing : int, optional
            Number of iteration over which to average the metric values
        iter_per_epoch : float, optional
            Number of iterations to complete an dataset epoch
        same_plot : bool, default True
            If `True`, draw all metrics on the same plot
        leg_ncols : int, default 1
            Number of columns in the legend
        figure_name : str, optional
            Name of the figure. If specified, figure is saved
        """
        # If the model/metric is a single string, nest it
        if isinstance(model, str):
            model = [model]
        if isinstance(metric, str):
            metric = [metric]

        # Make sure each model is given a name
        model_name = model_name or {}
        if isinstance(model_name, str):
            assert len(model) == 1, (
                "Should provide a single `model_name` if there is a "
                "single `model` to be represented."
            )
            model_name = {model[0]: model_name}
        else:
            for m in model:
                if m not in model_name:
                    model_name[m] = m

        # Make sure each metric is given a name
        metric_name = metric_name or {}
        if isinstance(metric_name, str):
            assert len(metric) == 1, (
                "Should provide a single `metric_name` if there is a "
                "single `metric` to be represented."
            )
            metric_name = {metric[0]: metric_name}
        else:
            for m in metric:
                if m not in metric_name:
                    metric_name[m] = m.split(self.separator)[0]

        # Make sure that each metric is given a range, if specified
        if limits is not None and not isinstance(limits, dict):
            limits = {m: limits for m in metric}

        # If using plotly, reset the layout and the trace list
        if self.interactive:
            self.initialize_plotly()
            traces = []

        # Initialize the figure to draw the training/validation metrics on
        if not same_plot and len(metric) > 1:
            if not self.interactive:
                # Prepare the matplotlib sublots
                fig, axes = plt.subplots(len(metric), sharex=True)
                fig.subplots_adjust(hspace=0)

                # Set all the axes with the same color
                for axis in axes:
                    axis.set_facecolor("white")

            else:
                # Prepare the plotly subplots
                fig = psubplots.make_subplots(
                    rows=len(metric), shared_xaxes=True, vertical_spacing=0
                )

                # Loop over the metrics
                for i, m in enumerate(metric):
                    # Set all axes formatting the same
                    if i > 0:
                        self.layout[f"xaxis{i+1}"] = self.layout["xaxis"]
                        self.layout[f"yaxis{i+1}"] = self.layout["yaxis"]

                    # Set all x axis empty apart from the bottom one
                    xtitle = None if i != len(metric) - 1 else "Epochs"
                    self.layout[f"xaxis{i+1}"]["title"]["text"] = xtitle

                    # Set all y axis the same
                    self.layout[f"yaxis{i+1}"]["title"]["text"] = metric_name[m]

                    # If there are limits specified,
                    if limits is not None and m in limits:
                        self.layout[f"yaxis{i+1}"]["range"] = limits[m]

                fig.update_layout(self.layout)

        elif self.interactive:
            # Set the limit for the shared y axis
            if limits is not None and metric[0] in limits:
                self.layout["yaxis"]["range"] = limits[metric[0]]

            # If there is a single metric, rename the y axis with its name
            if len(metric) == 1:
                self.layout["yaxis"]["title"]["text"] = metric_name[metric[0]]

            # If all the metrics are on a shared plot, give the legend a title
            if same_plot and len(model) == 1:
                self.layout["legend"]["title"] = model_name[model[0]]

            # Initialize the plotly figure
            fig = go.Figure(layout=self.layout)

        # Get the DataFrames for the requested models/metrics
        dfs, val_dfs, colors, draw_val = {}, {}, {}, {}
        for i, key in enumerate(model):
            log_subdir = self.log_dir + key
            dfs[key] = self.get_training_df(log_subdir, metric)
            val_dfs[key] = self.get_validation_df(log_subdir, metric)
            draw_val[key] = bool(len(val_dfs[key]["iter"]))
            colors[key] = self.colors[i % len(self.colors)]

        # Loop over the requested metrics, append a trace
        for i, metric_list in enumerate(metric):
            # Get a trace per training campaign
            for j, key in enumerate(dfs.keys()):
                # Get the necessary data
                iter_t, epoch_t = dfs[key]["iter"], dfs[key]["epoch"]
                metric_key, metric_label = self.find_key(dfs[key], metric_list)
                metric_t = dfs[key][metric_key]

                # If validation points are available, fetch the validation data
                if draw_val[key]:
                    # Get the necessary data
                    iter_v = val_dfs[key]["iter"]
                    metric_v_mean = val_dfs[key][metric_label + "_mean"]
                    metric_v_err = val_dfs[key][metric_label + "_err"]

                    # Convert the iteration number of each file to an epoch value
                    if iter_per_epoch is None:
                        epoch_v = [epoch_t[iter_t == it] for it in iter_v]
                        mask = np.where(np.array([len(e) for e in epoch_v]) == 1)[0]
                        epoch_v = [float(epoch_v[i].iloc[0]) for i in mask]
                        iter_v = iter_v[mask]
                        metric_v_mean = metric_v_mean[mask]
                        metric_v_err = metric_v_err[mask]
                    else:
                        epoch_v = iter_v / iter_per_epoch

                    # If requested, restrict the number of points to a range
                    if max_iter is not None:
                        mask_val = np.where(iter_v < max_iter)[0]
                        iter_v = iter_v[mask_val]
                        epoch_v = epoch_v[mask_val]
                        metric_v_mean = metric_v_mean[mask_val]
                        metric_v_err = metric_v_err[mask_val]

                # If requested, restrict the data to a certain range
                if max_iter is not None:
                    epoch_t = epoch_t[:max_iter]
                    metric_t = metric_t[:max_iter]

                # If requested, smooth out the curve by taking a rolling average
                if smoothing is not None and smoothing > 1:
                    metric_t = metric_t.rolling(
                        smoothing, min_periods=1, center=True
                    ).mean()

                # If requested, do not display all the iterations
                if step is not None and step > 1:
                    epoch_t = epoch_t[::step]
                    metric_t = metric_t[::step]

                # Pick a label for this specific model/metric pair
                if not same_plot:
                    label = model_name[key]
                else:
                    if len(model) == 1:
                        label = metric_name[metric_list]
                    elif len(metric) == 1:
                        label = model_name[key]
                    else:
                        label = f"{metric_name[metric_list]} ({model_name[key]})"
                    if print_min and draw_val[key]:
                        min_it = iter_v[np.argmin(metric_v_mean)]
                        label += f"{self.cr_char}Min: {min_it:d}"
                    if print_max and draw_val[key]:
                        max_it = iter_v[np.argmax(metric_v_mean)]
                        label += f"{self.cr_char}Max: {max_it:d}"

                # Fetch the appropriate color
                idx = i * len(model) + j
                if same_plot:
                    color = self.colors[idx % len(self.colors)]
                else:
                    color = colors[key]

                # Prepare the traces
                if not self.interactive:
                    # Add a trace for the training curve
                    axis = plt if same_plot else axes[i]
                    axis.plot(
                        epoch_t,
                        metric_t,
                        label=label,
                        color=color,
                        alpha=self.alpha,
                        linewidth=self.linewidth,
                    )

                    # Add a trace for the validation points
                    if draw_val[key]:
                        axis.errorbar(
                            epoch_v,
                            metric_v_mean,
                            yerr=metric_v_err,
                            fmt=".",
                            color=color,
                            linewidth=self.linewidth,
                            markersize=self.markersize,
                        )

                else:
                    # Add a trace for the training curve
                    legendgroup = f"group{idx}"
                    showlegend = same_plot or not i
                    traces += [
                        go.Scatter(
                            x=epoch_t,
                            y=metric_t,
                            name=label,
                            line={"color": color},
                            legendgroup=legendgroup,
                            showlegend=showlegend,
                        )
                    ]

                    # Add a trace for the validation points
                    if draw_val[key]:
                        hovertext = [f"(Iteration: {it:d})" for it in iter_v]
                        traces += [
                            go.Scatter(
                                x=epoch_v,
                                y=metric_v_mean,
                                error_y_array=metric_v_err,
                                mode="markers",
                                hovertext=hovertext,
                                marker={"color": color},
                                legendgroup=legendgroup,
                                showlegend=False,
                            )
                        ]

        if not self.interactive:
            if not same_plot:
                for i, metric_key in enumerate(metric):
                    axes[i].set_xlabel("Epochs")
                    axes[i].set_ylabel(metric_name[metric_key])
                    if limits is not None and metric_key in limits:
                        axes[i].set_ylim(limits[metric_key])
                axes[0].legend(ncol=leg_ncols)

            else:
                plt.xlabel("Epochs")
                ylabel = metric_name[metric[0]]
                plt.ylabel(ylabel if len(metric) == 1 else "Metric")
                plt.gca().set_ylim(limits[metric[0]])
                legend_title = model_name[model[0]] if len(model) == 1 else None
                plt.legend(
                    ncol=leg_ncols,
                    title=legend_title,
                    bbox_to_anchor=(1.05, 1),
                    loc="upper left",
                )

            if figure_name:
                plt.savefig(f"{figure_name}.png", bbox_inches="tight")
                plt.savefig(f"{figure_name}.pdf", bbox_inches="tight")

            plt.show()

        else:
            if not same_plot:
                n_mod, n_met = len(model), len(metric)
                mult = np.sum([2**val for val in draw_val.values()])
                step = 1.0 / mult
                rows = list(np.arange(n_met, step=step).astype(int) + 1)
                cols = list(np.ones(mult * n_met, dtype=int))
                fig.add_traces(traces, rows=rows, cols=cols)

            else:
                fig.add_traces(traces)

            iplot(fig)

    def get_training_df(self, log_dir, keys):
        """Finds all training log files inside the specified directory and
        concatenates them. If the range of iterations overlap, keep only that
        from the file started further in the training.

        Assumes that the formatting of the log file names is of the form
        `self.train_prefix-x.csv`, with `x` the number of iterations.

        Parameters
        ----------
        log_dir : str
            Path to the directory that contains the training log files
        keys : List[str]
            List of quantities to extract from the log files

        Returns
        -------
        pd.DataFrame
            Combined training log data
        """
        # Get all the log files that fit the pattern
        log_files = glob.glob(f"{log_dir}/{self.train_prefix}*")
        if not log_files:
            raise FileNotFoundError(
                f"Found no train log with prefix '{self.train_prefix}' "
                f"under {log_dir}."
            )

        # Find the iteration at which each successsive log file picks up to
        # avoid duplicate values at a given iteration
        start_iter = np.zeros(len(log_files), dtype=np.int64)
        for i, f in enumerate(log_files):
            start_iter[i] = int(f.split("-")[-1].split(".csv")[0])
        order = np.argsort(start_iter)
        end_points = np.append(start_iter[order], 1e12)

        # Loop over the log files, concatenate them
        log_dfs = []
        for i, f in enumerate(np.array(log_files)[order]):
            df = pd.read_csv(f, nrows=end_points[i + 1] - end_points[i])
            if len(df) == 0:
                continue
            for key_list in keys:
                key, key_name = self.find_key(df, key_list)
                df[key_name] = df[key]
            log_dfs.append(df)

        return pd.concat(log_dfs, sort=True)

    def get_validation_df(self, log_dir, keys):
        """Finds all validation log files inside the specified directory and
        build a single dataframe out of them. It returns the mean and std of
        the requested keys for each file.

        Assumes that the formatting of the log file names is of the form
        `self.val_prefix-x.csv`, with `x` the number of iterations.

        The key list allows for `:`-separated names, in case separate files
        use different names for the same quantity.

        Parameters
        ----------
        log_dir : str
            Path to the directory that contains the validation log files
        keys : List[str]
            List of quantities to extract from the log files

        Returns
        -------
        pd.DataFrame
            Combined validation log data
        """
        # Initialize a dictionary
        val_data = {"iter": []}
        for key in keys:
            key_name = key.split(":")[0]
            val_data[key_name + "_mean"] = []
            val_data[key_name + "_err"] = []

        # Loop over validation log files
        log_files = np.array(glob.glob(f"{log_dir}/{self.val_prefix}*"))
        for log_file in log_files:
            df = pd.read_csv(log_file)
            it = int(log_file.split("/")[-1].split("-")[-1].split(".")[0])
            val_data["iter"].append(it - 1)
            for key_list in keys:
                key, key_name = self.find_key(df, key_list)
                mean, err = df[key].mean(), df[key].std() / np.sqrt(len(df[key]))
                val_data[f"{key_name}_mean"].append(mean)
                val_data[f"{key_name}_err"].append(err)

        args = np.argsort(val_data["iter"])
        for key, val in val_data.items():
            val_data[key] = np.array(val)[args]

        return pd.DataFrame(val_data)

    def find_key(self, df, key_list):
        """Checks if a :class:`pd.DataFrame` contains any of the keys listed
        in a character-separated string.

        If multiple keys match, pick the first one.

        Parameters
        ----------
        df : Union[pd.DataFrame, dict]
            Pandas dataframe or dictionary containing data
        key_list : str
            Character-separated list of acceptable names for an attribute

        Returns
        -------
        key : str
            First data key which matches one of the keys in the list
        key_name : str
            First key in the list of acceptale names
        """
        key_list = key_list.split(self.separator)
        key_name = key_list[0]
        key_found = np.array([k in df.keys() for k in key_list])
        if not np.any(key_found):
            raise KeyError("Could not find any of the keys provided:", key_list)
        key = key_list[np.where(key_found)[0][0]]

        return key, key_name



"""Module with classes for all reconstructed and true objects."""

from dataclasses import dataclass, field

import numpy as np

from spine.data.base import PosDataBase
from spine.utils.docstring import inherit_docstring


@dataclass(eq=False)
class OutBase(PosDataBase):
    """Base data structure shared among all output classes.

    Attributes
    ----------
    id : int
        Unique index of the object within the object list
    index : np.ndarray
        (N) Voxel indexes corresponding to this object in the input tensor
    size : int
        Number of points, N, that make up this object
    points : np.ndarray
        (N, 3) Set of voxel coordinates that make up this object
    depositions : np.ndarray
        (N) Array of charge deposition values for each voxel
    depositions_sum : float
        Total amount of depositions
    sources : np.ndarray
        (N, 2) Set of voxel sources as (Module ID, TPC ID) pairs
    module_ids : np.ndarray
        (M) List of unique module indexes that make up this object
    is_contained : bool
        Whether this object is fully contained within the detector
    is_time_contained : bool
        Whether this object's points are within the expected readout window
    is_matched: bool
        True if a true object match was found
    match_ids : np.ndarray
        List of true object IDs this object is matched to
    match_overlaps : np.ndarray
        List of match overlaps (IoU) between the object and its matches
    is_cathode_crosser : bool
        True if the particle crossed a cathode, i.e. if it is made up
        of space points coming from > 1 TPC in one module
    cathode_offset : float
        If the particle is a cathode crosser, this is how far in cm one needs to
        move its points along the drift direction to reconcile at the cathode.
        This is directly proportional to time through time=offset/vdrift
    is_truth: bool
        Whether this object contains truth information or not
    units : str
        Units in which coordinates are expressed
    """

    id: int = -1
    index: np.ndarray = None
    size: int = None
    points: np.ndarray = None
    depositions: np.ndarray = None
    depositions_sum: float = None
    sources: np.ndarray = None
    module_ids: np.ndarray = None
    is_contained: bool = False
    is_time_contained: bool = False
    is_matched: bool = False
    match_ids: np.ndarray = None
    match_overlaps: np.ndarray = None
    is_cathode_crosser: bool = False
    cathode_offset: float = -np.inf
    is_truth: bool = None
    units: str = "cm"

    # Variable-length attribtues
    _var_length_attrs = (
        ("index", np.int64),
        ("depositions", np.float32),
        ("match_ids", np.int64),
        ("match_overlaps", np.float32),
        ("points", (3, np.float32)),
        ("sources", (2, np.int64)),
        ("module_ids", np.int64),
    )

    # Boolean attributes
    _bool_attrs = (
        "is_contained",
        "is_time_contained",
        "is_matched",
        "is_cathode_crosser",
        "is_truth",
    )

    # Attributes to concatenate when merging objects
    _cat_attrs = ("index", "points", "depositions", "sources")

    # Attributes that must never be stored to file
    _skip_attrs = ("points", "depositions", "sources")

    # Attributes that must not be stored to file when storing lite files
    _lite_skip_attrs = ("index",)

    def reset_match(self):
        """Resets the reco/truth matching information for the object."""
        self.is_matched = False
        self.match_ids = np.empty(0, dtype=np.int64)

    def reset_cathode_crosser(self):
        """Resets the cathode crossing information for the object."""
        self.is_cathode_crosser = False
        self.cathode_offset = -np.inf

    @property
    def size(self):
        """Total number of voxels that make up the object.

        Returns
        -------
        int
            Total number of voxels in the object
        """
        return len(self.index)

    @size.setter
    def size(self, size):
        pass

    @property
    def depositions_sum(self):
        """Total deposition value for the entire object.

        Returns
        -------
        float
            Sum of all depositions that make up the object
        """
        return np.sum(self.depositions)

    @depositions_sum.setter
    def depositions_sum(self, depositions_sum):
        pass

    @property
    def module_ids(self):
        """List of modules that contribute to this object.

        Returns
        -------
        np.ndarray
            List of unique modules contributing to this object.
        """
        return np.unique(self.sources[:, 0])

    @module_ids.setter
    def module_ids(self, module_ids):
        pass


@dataclass(eq=False)
@inherit_docstring(OutBase)
class RecoBase(OutBase):
    """Base data structure shared among all reconstructed output classes."""

    is_truth: bool = False


@dataclass(eq=False)
@inherit_docstring(OutBase)
class TruthBase(OutBase):
    """Base data structure shared among all truth output classes.

    Attributes
    ----------
    orig_id : int
        If matched to an MC truth instance, ID of the original instance
    depositions_q : np.ndarray
        (N) Array of values for each voxel in the same units as the input image
    depositions_q_sum : float
        Total amount of depositions in the same units as the input image
    index_adapt: np.ndarray
        (N') Voxel indexes corresponding to this object in the adapted cluster
        label tensor
    size_adapt : int
        Number of points, N', that make up this object in the adapted cluster
        label tensor
    points_adapt : np.ndarray
        (N', 3) Set of voxel coordinates using adapted cluster labels
    depositions_adapt : np.ndarray
        (N') Array of values for each voxel in the adapted cluster label tensor
    depositions_adapt_sum : float
        Total amount of depositions in adapted cluster label tensor
    depositions_adapt_q : np.ndarray
        (N) Array of values for each voxel in the same units as the input image
    depositions_adapt_q_sum : float
        Total amount of depositions in adapted cluster label tensor in the same
        units as the input image
    sources_adapt : np.ndarray
        (N, 2) Set of voxel sources as (Module ID, TPC ID) pairs, adapted
    index_g4: np.ndarray
        (N'') Fragment voxel indexes in the true Geant4 energy deposition tensor
    size_g4 : int
        Number of points, N'', that make up this object in the true Geant4
        energy deposition tensor
    points_g4 : np.ndarray
        (N'', 3) Set of voxel coordinates of true Geant4 energy depositions
    depositions_g4 : np.ndarray
        (N'') Array of true Geant4 energy depositions per voxel
    depositions_g4_sum : float
        Total amount of true Geant4 depositions
    """

    orig_id: int = -1
    depositions_q: np.ndarray = None
    depositions_q_sum: float = None
    index_adapt: np.ndarray = None
    size_adapt: int = None
    size_g4: int = None
    points_adapt: np.ndarray = None
    depositions_adapt: np.ndarray = None
    depositions_adapt_sum: float = None
    depositions_adapt_q: np.ndarray = None
    depositions_adapt_q_sum: float = None
    sources_adapt: np.ndarray = None
    index_g4: np.ndarray = None
    points_g4: np.ndarray = None
    depositions_g4: np.ndarray = None
    depositions_g4_sum: float = None
    is_truth: bool = True

    # Variable-length attribtues
    _var_length_attrs = (
        ("depositions_q", np.float32),
        ("index_adapt", np.int64),
        ("depositions_adapt", np.float32),
        ("depositions_adapt_q", np.float32),
        ("index_g4", np.int64),
        ("depositions_g4", np.float32),
        ("points_adapt", (3, np.float32)),
        ("sources_adapt", (2, np.int64)),
        ("points_g4", (3, np.float32)),
        *OutBase._var_length_attrs,
    )

    # Attributes to concatenate when merging objects
    _cat_attrs = (
        "depositions_q",
        "index_adapt",
        "points_adapt",
        "depositions_adapt",
        "depositions_adapt_q",
        "sources_adapt",
        "index_g4",
        "points_g4",
        "depositions_g4",
        *OutBase._cat_attrs,
    )

    # Attributes that must never be stored to file
    _skip_attrs = (
        "depositions_q",
        "points_adapt",
        "depositions_adapt",
        "depositions_adapt_q",
        "sources_adapt",
        "depositions_g4",
        "points_g4",
        "depositions_g4",
        *OutBase._skip_attrs,
    )

    # Attributes that must not be stored to file when storing lite files
    _lite_skip_attrs = ("index_adapt", "index_g4", *OutBase._lite_skip_attrs)

    @property
    def size_adapt(self):
        """Total number of voxels that make up the object in the adapted tensor.

        Returns
        -------
        int
            Total number of voxels in the object
        """
        return len(self.index_adapt)

    @size_adapt.setter
    def size_adapt(self, size_adapt):
        pass

    @property
    def size_g4(self):
        """Total number of voxels that make up the object in the Geant4 tensor.

        Returns
        -------
        int
            Total number of voxels in the object
        """
        return len(self.index_g4)

    @size_g4.setter
    def size_g4(self, size_g4):
        pass

    @property
    def depositions_q_sum(self):
        """Total deposition value for the entire object in the original units.

        Returns
        -------
        float
            Sum of all depositions that make up the object
        """
        return np.sum(self.depositions_q)

    @depositions_q_sum.setter
    def depositions_q_sum(self, depositions_q_sum):
        pass

    @property
    def depositions_adapt_sum(self):
        """Total deposition value for the entire object in the adapted tensor.

        Returns
        -------
        float
            Sum of all depositions that make up the object
        """
        return np.sum(self.depositions_adapt)

    @depositions_adapt_sum.setter
    def depositions_adapt_sum(self, depositions_adapt_sum):
        pass

    @property
    def depositions_adapt_q_sum(self):
        """Total deposition value for the entire object in the adapted tensor
        and in the original units.

        Returns
        -------
        float
            Sum of all depositions that make up the object
        """
        return np.sum(self.depositions_adapt_q)

    @depositions_adapt_q_sum.setter
    def depositions_adapt_q_sum(self, depositions_adapt_q_sum):
        pass

    @property
    def depositions_g4_sum(self):
        """Total deposition value for the entire object in the Geant4 tensor.

        Returns
        -------
        float
            Sum of all depositions that make up the object
        """
        return np.sum(self.depositions_g4)

    @depositions_g4_sum.setter
    def depositions_g4_sum(self, depositions_g4_sum):
        pass



"""Contains base class of all post-processors."""

from abc import ABC, abstractmethod


class PostBase(ABC):
    """Base class of all post-processors.

    This base class performs the following functions:
      - Ensures that the necessary method exist
      - Checks that the post-processor is provided the necessary information
        to do its job
      - Fetches the appropriate coordinate attributes
      - Ensures that the appropriate units are provided

    Attributes
    ----------
    name : str
        Name of the post-processor as defined in the configuration file
    aliases : Tuple[str]
        Alternative acceptable names for a post-processor
    """

    # Name of the post-processor (as specified in the configuration)
    name = None

    # Alternative allowed names of the post-processor
    aliases = ()

    # Units in which the post-processor expects objects to be expressed in
    units = "cm"

    # Whether this post-processor needs to know where the configuration lives
    need_parent_path = False

    # Set of data keys needed for this post-processor to operate
    _keys = ()

    # Set of post-processors which must be run before this one is
    _upstream = ()

    # List of recognized object types
    _obj_types = ("fragment", "particle", "interaction")

    # List of recognized run modes
    _run_modes = ("reco", "truth", "both", "all")

    # List of known reconstructed particle identification modes
    _pid_modes = ("pid", "chi2_pid")

    # List of known point modes for true particles and their corresponding keys
    _point_modes = (
        ("points", "points_label"),
        ("points_adapt", "points"),
        ("points_g4", "points_g4"),
    )

    # List of known source modes for true particles and their corresponding keys
    _source_modes = (
        ("sources", "sources_label"),
        ("sources_adapt", "sources"),
        ("sources_g4", "sources_g4"),
    )

    # List of known deposition modes for true particles and their corresponding keys
    _dep_modes = (
        ("depositions", "depositions_label"),
        ("depositions_q", "depositions_q_label"),
        ("depositions_adapt", "depositions_label_adapt"),
        ("depositions_adapt_q", "depositions"),
        ("depositions_g4", "depositions_g4"),
    )

    def __init__(
        self,
        obj_type=None,
        run_mode=None,
        truth_point_mode=None,
        truth_dep_mode=None,
        pid_mode=None,
        parent_path=None,
    ):
        """Initialize default post-processor object properties.

        Parameters
        ----------
        obj_type : Union[str, List[str]]
            Name or list of names of the object types to process
        run_mode : str, optional
            If specified, tells whether the post-processor must run on
            reconstructed ('reco'), true ('true') or both objects
            ('both' or 'all')
        truth_point_mode : str, optional
            If specified, tells which attribute of the :class:`TruthFragment`,
            :class:`TruthParticle` or :class:`TruthInteraction` object to use
            to fetch its point coordinates
        truth_dep_mode : str, optional
            If specified, tells which attribute of the :class:`TruthFragment`,
            :class:`TruthParticle` or :class:`TruthInteraction` object to use
            to fetch its depositions
        parent_path : str, optional
            Path to the parent directory of the main analysis configuration. This
            allows for the use of relative paths in the post-processors.
        """
        # If run mode is specified, process it
        if run_mode is not None:
            # Check that the run mode is recognized
            assert run_mode in self._run_modes, (
                f"`run_mode` not recognized: {run_mode}. Must be one of "
                f"{self._run_modes}."
            )

        # Check that all the object sources are recognized
        if obj_type is None:
            obj_type = []
        elif isinstance(obj_type, str):
            obj_type = [obj_type]

        for obj in obj_type:
            assert obj in self._obj_types, (
                f"Object type must be one of {self._obj_types}. Got "
                f"`{obj}` instead."
            )

        # Make a list of object keys to process
        self.fragment_keys = []
        self.particle_keys = []
        self.interaction_keys = []
        for name in self._obj_types:
            # Initialize one list per object type
            setattr(self, f"{name}_keys", [])

            # Skip object types which are not requested
            if name in obj_type:
                if run_mode != "truth":
                    getattr(self, f"{name}_keys").append(f"reco_{name}s")
                if run_mode != "reco":
                    getattr(self, f"{name}_keys").append(f"truth_{name}s")

        self.obj_keys = self.fragment_keys + self.particle_keys + self.interaction_keys

        # Update underlying keys, if needed
        self.update_keys({k: True for k in self.obj_keys})

        # If a truth point mode is specified, store it
        if truth_point_mode is not None:
            assert truth_point_mode in self.point_modes, (
                "The `truth_point_mode` argument must be one of "
                f"{self.point_modes.keys()}. Got `{truth_point_mode}` instead."
            )
            self.truth_point_mode = truth_point_mode
            self.truth_point_key = self.point_modes[self.truth_point_mode]
            self.truth_source_mode = truth_point_mode.replace("points", "sources")
            self.truth_source_key = self.source_modes[self.truth_source_mode]
            self.truth_index_mode = truth_point_mode.replace("points", "index")

        # If a truth deposition mode is specified, store it
        if truth_dep_mode is not None:
            assert truth_dep_mode in self.dep_modes, (
                "The `truth_dep_mode` argument must be one of "
                f"{self.dep_modes.keys()}. Got `{truth_dep_mode}` instead."
            )
            if truth_point_mode is not None:
                prefix = truth_point_mode.replace("points", "depositions")
                assert truth_dep_mode.startswith(prefix), (
                    f"Points mode {truth_point_mode} and deposition mode "
                    f"{truth_dep_mode} are incompatible."
                )
            self.truth_dep_mode = truth_dep_mode
            self.truth_dep_key = self.dep_modes[truth_dep_mode]

        # If a PID mode is specified, store it
        if pid_mode is not None:
            assert pid_mode in self._pid_modes, (
                f"The `pid_mode` argument must be one of {self._pid_modes}. "
                f"Got {pid_mode} instead."
            )
            self.pid_mode = pid_mode

        # Store the parent path
        self.parent_path = parent_path
        if self.need_parent_path and parent_path is None:
            raise ValueError(
                f"The `{self.name}` post-processor needs to know the parent "
                "configuration path. Please provide the `parent_path` argument."
            )

    @property
    def keys(self):
        """Dictionary of (key, necessity) pairs which determine which data keys
        are needed/optional for the post-processor to run.

        Returns
        -------
        Dict[str, bool]
            Dictionary of (key, necessity) pairs to be used
        """
        return dict(self._keys)

    @property
    def point_modes(self):
        """Dictionary which makes the correspondance between the name of a true
        object point attribute with the underlying point tensor it points to.

        Returns
        -------
        Dict[str, str]
            Dictionary of (attribute, key) mapping for point coordinates
        """
        return dict(self._point_modes)

    @property
    def source_modes(self):
        """Dictionary which makes the correspondance between the name of a true
        object source attribute with the underlying source tensor it points to.

        Returns
        -------
        Dict[str, str]
            Dictionary of (attribute, key) mapping for point sources
        """
        return dict(self._source_modes)

    @property
    def dep_modes(self):
        """Dictionary which makes the correspondance between the name of a true
        object deposition attribute with the underlying deposition array it points to.

        Returns
        -------
        Dict[str, str]
            Dictionary of (attribute, key) mapping for point depositions
        """
        return dict(self._dep_modes)

    def update_keys(self, update_dict):
        """Update the underlying set of keys and their necessity in place.

        Parameters
        ----------
        update_dict : Dict[str, bool]
            Dictionary of (key, necessity) pairs to update the keys with
        """
        if len(update_dict) > 0:
            keys = self.keys
            keys.update(update_dict)
            self._keys = tuple(keys.items())

    def update_upstream(self, key):
        """Update the underlying set of required upstream modules in place.

        Parameters
        ----------
        key : str
            Post-processor module name to add to the list
        """
        self._upstream = (*self._upstream, key)

    def __call__(self, data, entry=None):
        """Calls the post processor on one entry.

        Parameters
        ----------
        data : dict
            Dicitionary of data products
        entry : int, optional
            Entry in the batch

        Returns
        -------
        dict
            Update to the input dictionary
        """
        # Fetch the input dictionary
        data_filter = {}
        for key, req in self._keys:
            # If this key is needed, check that it exists
            assert not req or key in data, (
                f"Post-processor `{self.name}` is missing an essential "
                f"input to be used: `{key}`."
            )

            # Append
            if key in data:
                data_filter[key] = data[key]
                if entry is not None:
                    data_filter[key] = data[key][entry]

        # Run the post-processor
        return self.process(data_filter)

    def get_index(self, obj):
        """Get a certain pre-defined index attribute of an object.

        The :class:`TruthFragment`, :class:`TruthParticle` and
        :class:`TruthInteraction` objects index are obtained using the
        `truth_index_mode` attribute of the class.

        Parameters
        ----------
        obj : Union[FragmentBase, ParticleBase, InteractionBase]
            Fragment, Particle or Interaction object

        Results
        -------
        np.ndarray
            (N) Object index
        """
        if not obj.is_truth:
            return obj.index
        else:
            return getattr(obj, self.truth_index_mode)

    def get_points(self, obj):
        """Get a certain pre-defined point attribute of an object.

        The :class:`TruthFragment`, :class:`TruthParticle` and
        :class:`TruthInteraction` objects points are obtained using the
        `truth_point_mode` attribute of the class.

        Parameters
        ----------
        obj : Union[FragmentBase, ParticleBase, InteractionBase]
            Fragment, Particle or Interaction object

        Results
        -------
        np.ndarray
            (N, 3) Point coordinates
        """
        if not obj.is_truth:
            return obj.points
        else:
            return getattr(obj, self.truth_point_mode)

    def get_sources(self, obj):
        """Get a certain pre-defined sources attribute of an object.

        The :class:`TruthFragment`, :class:`TruthParticle` and
        :class:`TruthInteraction` objects sources are obtained using the
        `truth_source_mode` attribute of the class.

        Parameters
        ----------
        obj : Union[FragmentBase, ParticleBase, InteractionBase]
            Fragment, Particle or Interaction object

        Results
        -------
        np.ndarray
            (N, 2) Object sources
        """
        if not obj.is_truth:
            return obj.sources
        else:
            return getattr(obj, self.truth_source_mode)

    def get_depositions(self, obj):
        """Get a certain pre-defined deposition attribute of an object.

        The :class:`TruthFragment`, :class:`TruthParticle` and
        :class:`TruthInteraction` objects points are obtained using the
        `truth_dep_mode` attribute of the class.

        Parameters
        ----------
        obj : Union[FragmentBase, ParticleBase, InteractionBase]
            Fragment, Particle or Interaction object

        Results
        -------
        np.ndarray
            (N) Depositions
        """
        if not obj.is_truth:
            return obj.depositions
        else:
            return getattr(obj, self.truth_dep_mode)

    def get_pid(self, obj):
        """Get a certain pre-defined PID prediction of an object.

        The :class:`TruthParticle` PID predictions are obtained using the
        `pid_mode` attribute of the class.

        Parameters
        ----------
        obj : Union[ParticleBase]
            Particle object

        Results
        -------
        int
            Particle identification enumerator
        """
        if not obj.is_truth:
            return getattr(obj, self.pid_mode)
        else:
            return obj.pid

    def check_units(self, obj):
        """Check that the point coordinates of an object are as expected.

        Parameters
        ----------
        obj : Union[FragmentBase, ParticleBase, InteractionBase]
            Particle or interaction object

        Results
        -------
        np.ndarray
            (N, 3) Point coordinates
        """
        if obj.units != self.units:
            raise ValueError(
                f"Coordinates must be expressed in {self.units} but are "
                f"currently in {obj.units} instead."
            )

    @abstractmethod
    def process(self, data):
        """Place-holder method to be defined in each post-processor.

        Parameters
        ----------
        data : dict
            Dictionary of processed data products
        """
        raise NotImplementedError("Must define the `process` function.")



"""Functions which initialize activation and normalization layers
used by CNNs. These factories build activations and normalizations layers
are based on the MinkowskiEngine package.
"""

from copy import deepcopy

from spine.utils.factory import instantiate

__all__ = ["act_factory", "norm_factory"]


def act_dict():
    """Dictionary of valid activation functions."""
    import MinkowskiEngine as ME
    import MinkowskiEngine.MinkowskiNonlinearity as MENL
    from torch.nn import Identity

    from . import nonlinearities

    activations = {
        "none": Identity,
        "relu": ME.MinkowskiReLU,
        "prelu": ME.MinkowskiPReLU,
        "selu": ME.MinkowskiSELU,
        "celu": ME.MinkowskiCELU,
        "tanh": ME.MinkowskiTanh,
        "sigmoid": ME.MinkowskiSigmoid,
        "lrelu": MENL.MinkowskiLeakyReLU,
        "elu": MENL.MinkowskiELU,
        "mish": nonlinearities.MinkowskiMish,
    }

    return activations


def norm_dict():
    """Dictionary of valid normalization functions."""
    import MinkowskiEngine as ME
    from torch.nn import Identity

    from . import normalizations

    norm_layers = {
        "none": Identity,
        "batch_norm": ME.MinkowskiBatchNorm,
        "instance_norm": ME.MinkowskiInstanceNorm,
        "pixel_norm": normalizations.MinkowskiPixelNorm,
    }

    return norm_layers


def act_factory(cfg):
    """Instantiates an activation layer.

    Parameters
    ----------
    cfg : dict
        Activation layer configuration

    Return
    ------
    object
        Instantiated activation layer
    """
    return instantiate(act_dict(), cfg)


def norm_factory(cfg, num_features=None):
    """Instantiates a normalization layer.

    Parameters
    ----------
    cfg : dict
        Normalization layer configuration
    num_features : int
        Number of features to normalize

    Return
    ------
    object
        Instantiated normalization layer
    """
    return instantiate(norm_dict(), cfg, num_features=num_features)



"""Module with a class that leverages particle endpoints to do clustering."""

import numpy as np
import scipy
from scipy.spatial.distance import cdist

from spine.math.cluster import DBSCAN

__all__ = ["PointBreakClusterer"]


class PointBreakClusterer:
    """Leverages particles start/end point positions to break up instances
    of particles which touch (vertex, secondary interaction point, etc.).

    Two methods are supported: `masked_dbscan` and `closest_path`.

    The masked DBSCAN method proceeds as follows:
    - Break up the input point cloud into DBSCAN clusters
    - For each cluster, mask out regions around the particle end points
    - Run DBSCAN on the masked point cloud
    - Each new instance is a particle instance
    - Unlabeled, masked points are merged to the closest instance

    The closest path method proceeds as follows:
    - Break up the input point cloud into DBSCAN clusters
    - For each cluster, build a radius graph on its constituents
    - Find the closest graph paths for each pair of particle end points
    - The paths that belong to the minimum spanning tree are particles
    - Points from the cloud are assigned to a particle using their
      proximity to a particle path

    The latter only works on track clusters, not on EM showers.
    """

    def __init__(
        self,
        method="masked_dbscan",
        eps=1.8,
        min_samples=1,
        metric="euclidean",
        mask_radius=5.0,
    ):
        """Initialize the particle point-enhanced clustering algorithm.

        Parameters
        ----------
        method : str, default 'masked_dbscan'
            Clustering method
        eps : float, default 1.8
            The maximum distance between two samples for one to be considered
            as in the neighborhood of the other.
        min_samples : int, default 1
            The number of samples (or total weight) in a neighborhood for a
            point to be considered as a core point.
        metric : str, default 'euclidean'
            Metric used to compute the pair-wise distances between space points
        mask_radius : float, default 5.0
            Radius to mask around each particle point
        """
        # Store the attributes
        self.method = method
        self.eps = eps
        self.min_samples = min_samples
        self.metric = metric
        self.mask_radius = mask_radius

        # Initialize the DBSCAN algorithm
        self.dbscan = DBSCAN(eps=eps, min_samples=min_samples, metric=metric)

    def __call__(self, voxels, points, method=None):
        """Produce instance clusters using the point-enhanced method.

        Parameters
        ----------
        voxels : np.ndarray
            (N, 3) Set of voxel coordinates
        points : np.ndarray
            (P, 3) Set of particle end points
        method : str, optional
            Override the method defined in the initialzer

        Returns
        -------
        np.ndarray
            (N) Array of cluster assignments for each voxel in the input
        """
        method = method if method is not None else self.method
        if method == "masked_dbscan":
            return self.get_masked_dbscan_labels(voxels, points)

        elif method == "closest_path":
            return self.get_closest_path_labels(voxels, points)

        else:
            raise ValueError(
                "Point-enhanced clustering method not recognized: "
                f"{self.method}. Should be one of 'masked_dbscan' or "
                "'closest_path'."
            )

    def get_masked_dbscan_labels(self, voxels, points):
        """Produce instance clusters using the masked-DBSCAN method.

        Parameters
        ----------
        voxels : np.ndarray
            (N, 3) Set of voxel coordinates
        points : np.ndarray
            (P, 3) Set of particle end points

        Returns
        -------
        np.ndarray
            (N) Array of cluster assignments for each voxel in the input
        """
        # Find voxels above a threshold distance from any particle end point
        pair_mat = cdist(voxels, points, metric=self.metric)
        dist_mask = np.all((pair_mat > self.mask_radius), axis=1)

        # Form preliminary clusters on the entire set of voxels, loop over them
        labels = self.dbscan.fit_predict(voxels)
        for l in np.unique(labels):
            # Find the list of masked and unmasked voxels in this cluster
            group_mask = labels == l
            active_index = np.where(dist_mask & group_mask)[0]
            passive_index = np.where(~dist_mask & group_mask)[0]

            if len(active_index):
                # If there are active pixels, run DBSCAN on the masked voxels
                labels_c = self.dbscan.fit_predict(voxels[active_index])
                labels[active_index] = np.max(labels) + 1 + labels_c

                if len(passive_index):
                    # For masked pixel, matched to the closest cluster
                    dist_mat = cdist(
                        voxels[active_index], voxels[passive_index], metric=self.metric
                    )
                    argmins = np.argmin(dist_mat, axis=0)
                    labels[passive_index] = labels[active_index][argmins]

        return labels

    def get_closest_path_labels(self, voxels, points):
        """Produce instance clusters using the closest-path method

        Parameters
        ----------
        voxels : np.ndarray
            (N, 3) Set of voxel coordinates
        points : np.ndarray
            (P, 3) Set of particle end points

        Returns
        -------
        np.ndarray
            (N) Array of cluster assignments for each voxel in the input
        """
        # Precompute the pair-wise distances between voxels and points
        pair_mat = cdist(voxels, points, metric=self.metric)

        # Form preliminary clusters on the entire set of voxels, loop over them
        labels = self.dbscan.fit_predict(voxels)
        for l in np.unique(labels):
            # Restrict voxel set and point set to those in the group
            group_mask = labels == l
            point_mask = np.min(pair_mat[group_mask], axis=0) < self.eps
            point_ids = np.unique(
                np.argmin(pair_mat[np.ix_(group_mask, point_mask)], axis=0)
            )
            if len(point_ids) > 2:
                # Build a graph on the group voxels that respect the DBSCAN distance scale
                dist_mat = cdist(
                    voxels[group_mask], voxels[group_mask], metric=self.metric
                )
                graph = dist_mat * (dist_mat < self.eps)
                cs_graph = scipy.sparse.csr_matrix(graph)

                # Find the shortest between each of the breaking points, identify segments that minimize absolute excursion
                graph_mat, predecessors = scipy.sparse.csgraph.shortest_path(
                    csgraph=cs_graph, directed=False, return_predecessors=True
                )
                break_ix = np.ix_(point_ids, point_ids)
                chord_mat = dist_mat[break_ix]
                mst_mat = scipy.sparse.csgraph.minimum_spanning_tree(
                    graph_mat[break_ix] - chord_mat + 1e-6
                ).toarray()
                mst_edges = np.vstack(np.where(mst_mat > 0)).T

                # Construct graph paths along the tree
                paths = [[] for _ in range(len(mst_edges))]
                for i, e in enumerate(mst_edges):
                    k, l = point_ids[e]
                    paths[i].append(l)
                    while l != k:
                        l = predecessors[k, l]
                        paths[i].append(l)

                # Find the path closest to each of the voxels in the group. If a path does not improve reachability, remove
                mindists = np.vstack([np.min(graph_mat[:, p], axis=1) for p in paths])
                mst_tort = mst_mat[mst_mat > 0] / chord_mat[mst_mat > 0]  # tau - 1
                tort_ord = np.argsort(-mst_tort)
                least_r = np.max(
                    np.min(mindists, axis=0)
                )  # Least reachable point distance
                path_mask = np.ones(len(mst_tort), dtype=np.bool)
                for i in range(len(mst_tort)):
                    if np.sum(path_mask) == 1:
                        break
                    path_mask[tort_ord[i]] = False
                    reach = np.max(np.min(mindists[path_mask], axis=0))
                    if reach > least_r:
                        path_mask[tort_ord[i]] = True

                # Associate voxels with closest remaining path
                mindists = np.vstack(
                    [
                        np.min(
                            graph_mat[:, p[min(1, len(p) - 2) : max(len(p) - 1, 2)]],
                            axis=1,
                        )
                        for i, p in enumerate(paths)
                        if path_mask[i]
                    ]
                )
                sublabels = np.argmin(mindists, axis=0)
                labels[group_mask] = max(labels) + 1 + sublabels

        return np.unique(labels, return_inverse=True)[1]



import copy
import datetime
import json
import os
import sys
import time
from abc import ABC
from collections import OrderedDict
from typing import List

import numpy as np
import pandas as pd
import torch
import yaml

from spine.io.factories import loader_factory
from spine.main_funcs import cycle, process_config
from spine.model.experimental.hyperopt.utils import (
    UniformDistribution,
    construct_eval_func,
)
from spine.trainval import trainval
from spine.utils import utils

from .factories import *


def make_config(base_config: dict, parametrization: List[dict]) -> dict:
    """
    Use hyperparameter search domain specification dictionary to
    prepare new training config with model/optimizer hyperparameters
    changed accordingly.
    """
    for d in parametrization:
        search_and_replace(
            base_config,
        )
    raise NotImplementedError


def update_progress(progress, loss, acc):
    """
    Pure Python Progress Bar
    Author: Brian Khuu, rayryeng
    Reference: https://stackoverflow.com/questions/3160699/python-progress-bar
    """
    barLength = 20
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength * progress))
    text = "\rPercent: [{0}] {1:.4f}% {2}    loss {3:.4f} accuracy {4:.4f}".format(
        "#" * block + "-" * (barLength - block), progress * 100, status, loss, acc
    )
    sys.stdout.write(text)
    sys.stdout.flush()


def search_and_replace(config: dict, key: str, val: float):
    """
    Recursively parse config for all instances of (key : val) pairs
    and replace it with new value.
    """
    val = float(val)
    if isinstance(config.get(key, None), type(val)):
        config[key] = val
        return
    elif isinstance(config.get(key, None), dict):
        search_and_replace(config[key], key, val)
    else:
        current_keys = config.keys()
        for k in current_keys:
            if isinstance(config[k], dict):
                search_and_replace(config[k], key, val)


def process_domain_config(parametrization):
    """
    Process hyperparameter domain configuration dictionary
    into Ax-compatible form.
    """
    params = []

    for variable_name, var_domain in parametrization.items():

        d = {}
        d["name"] = variable_name
        d.update(var_domain)
        params.append(d)

    return params


def setup_parameter_list(parametrization):

    params_unravel = []

    for domain in parametrization:
        variable_name = domain["name"]
        bounds = domain["bounds"]
        num = domain["num"]
        log_scale = domain.get("log_scale", False)
        if log_scale:
            grid = np.logspace(bounds[0], bounds[1], num)
        else:
            grid = np.linspace(bounds[0], bounds[1], num)

        params_unravel[variable_name] = grid

    params_keys = params_unravel.keys()


class HyperparameterSearch(ABC):

    def __init__(self, cfg: dict, eval_func: str = "default"):

        self.base_config = cfg
        self.eval_func = construct_eval_func(
            cfg["hyperparameter_search"].get("eval_func", "default")
        )

        self.log_dir = cfg["hyperparameter_search"]["log_dir"]
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        self.output_optimal_json = cfg["hyperparameter_search"].get(
            "output_optimal_json", "optimal_params.json"
        )
        self.output_results = cfg["hyperparameter_search"].get(
            "output_results", "results.csv"
        )

        # Training Configurations

        self.train_config = cfg["hyperparameter_search"]["train"]
        self.val_config = cfg["hyperparameter_search"]["validation"]

        process_config(self.train_config)
        process_config(self.val_config)

        self.num_train = self.train_config.get("size", 1000)
        self.trainset_size = self.train_config.get("trainset_size", 80000)

        if self.train_config.get("random_sample", True):

            self.train_seed = self.train_config.get("seed", None)

        np.random.seed(self.train_seed)
        self.train_event_list = np.random.randint(0, self.trainset_size, self.num_train)

        self.train_io = loader_factory(
            self.train_config, event_list=self.train_event_list
        )
        self.train_io_iter = iter(cycle(self.train_io))
        self.train_num_iter = int(self.train_config["trainval"]["iterations"])

        # Validation Configurations

        if self.val_config.get("random_sample", True):

            self.val_seed = self.val_config.get("seed", None)

        self.num_val = self.val_config.get("size", 100)
        self.valset_size = self.val_config.get("valset_size", 20000)

        np.random.seed(self.val_seed)
        self.val_event_list = list(np.random.randint(0, self.valset_size, self.num_val))

        self.val_io = loader_factory(self.val_config, event_list=self.val_event_list)
        self.val_io_iter = iter(cycle(self.val_io))
        self.val_num_iter = self.val_config["trainval"]["iterations"]

        self.params_config = cfg["hyperparameter_search"]["domain"]
        self.optimal_params = {}

    def train_evaluate(self, sampled_params: dict):

        local_config = copy.deepcopy(self.train_config)

        for key, val in sampled_params.items():
            print(key, val, type(key))
            search_and_replace(local_config, key, val)

        trainer = trainval(local_config)
        trainer.initialize()
        t_iter = trainer._watch.time("iteration")
        t_io = trainer._watch.time("io")
        t_save = trainer._watch.time("save")
        t_net = trainer._watch.time(
            "train" if local_config["trainval"]["train"] else "forward"
        )

        iteration = 0

        if len(local_config["trainval"]["gpus"]) > 0:
            torch.cuda.set_device(local_config["trainval"]["gpus"][0])

        while iteration < self.train_num_iter:

            start = time.time()

            prog = float(iteration) / self.train_num_iter

            data_blob, result_blob = trainer.train_step(self.train_io_iter)

            acc = round(np.mean(result_blob.get("accuracy", -1)), 4)
            loss = round(np.mean(result_blob.get("loss", -1)), 4)

            end = time.time()
            tabs = end - start

            epoch = iteration / float(len(self.train_io))

            if torch.cuda.is_available():
                mem = round(torch.cuda.max_memory_allocated() / 1.0e9, 3)

            tstamp_iteration = datetime.datetime.fromtimestamp(time.time()).strftime(
                "%Y-%m-%d %H:%M:%S"
            )

            update_progress(prog, loss, acc)

            iteration += 1

        with torch.no_grad():
            accuracy = self.compute_accuracy(trainer)

        del trainer
        return accuracy

    def compute_accuracy(self, trainer) -> float:
        """
        Compute accuracy score for a single datapoint (one trained model)
        """
        accuracy = []
        self.eval_func(trainer._net)
        iteration = 0
        while iteration < self.val_num_iter:
            data_blob, result_blob = trainer.forward(self.val_io_iter)
            acc = result_blob["accuracy"][0]
            accuracy.append(acc)
            iteration += 1
        accuracy = sum(accuracy) / len(accuracy)
        return accuracy


class GridSearch(HyperparameterSearch):

    def __init__(self, cfg: dict, eval_func: str = "default"):
        super(GridSearch, self).__init__(cfg, eval_func)

    def optimize(self, parametrization):
        pass


class RandomSearch(HyperparameterSearch):

    def __init__(self, cfg: dict, eval_func: str = "default"):
        super(RandomSearch, self).__init__(cfg, eval_func)
        self.num_trials = cfg["hyperparameter_search"].get("num_trials", 30)
        self.set_samplers()

    def set_samplers(self):

        distributions = OrderedDict()

        for param, specs in self.params_config.items():
            lb, ub = specs["bounds"]
            log_scale = specs.get("log_scale", False)
            dist = UniformDistribution(lb, ub, log_scale=log_scale)
            distributions[param] = dist

        self.samplers = distributions

    def sample(self):

        gridpt = {}
        for param, dist in self.samplers.items():
            gridpt[param] = dist.sample()

        return gridpt

    def optimize_and_save(self):

        df = []
        max_acc = 0

        for i in range(self.num_trials):
            out_msg = "Running optimization trial {0}/{1}\n".format(i, self.num_trials)
            sys.stdout.write(out_msg)
            sys.stdout.flush()
            gridpt = self.sample()
            try:
                acc = self.train_evaluate(gridpt)
                gridpt.update({"accuracy": acc})
            except:
                # Training diverges, crashes, or other unexpected behavior
                gridpt.update({"accuracy": -1})
                acc = -1

            df.append(gridpt)
            # Check for maximum accuracy
            if max_acc < acc:
                max_acc = acc
                self.optimal_params = gridpt

            print("df = ", df)

        df = pd.DataFrame(df)
        self.log = df
        csv_path = os.path.join(self.log_dir, self.output_results)
        df.to_csv(csv_path)
        output_optimal_json = os.path.join(self.log_dir, self.output_optimal_json)
        with open(output_optimal_json, "w") as outfile:
            json.dump(self.optimal_params, outfile)
        return self.optimal_params


def search(config):
    hyperopt_config = config["hyperparameter_search"]
    name = hyperopt_config["name"]
    alg_constructor = construct_hyperopt_run(name)
    model = alg_constructor(config, hyperopt_config.get("eval_func", "default"))
    model.optimize_and_save()



import MinkowskiEngine as ME
import MinkowskiFunctional as MF
import numpy as np
import torch
import torch.nn as nn

from spine.model.layer.cnn.act_norm import act_factory
from spine.model.layer.cnn.blocks import ResNetBlock
from spine.model.layer.cnn.configuration import setup_cnn_configuration
from spine.model.layer.cnn.uresnet_layers import UResNetDecoder, UResNetEncoder


class Attention(nn.Module):
    """
    Sparse Attention Module where the feature map is multiplied
    by a soft masking score tensor (sigmoid activated)
    """

    def __init__(self):
        super(Attention, self).__init__()

    def forward(self, x, scores):
        features = x.F
        features = features * scores
        coords = x.C
        output = ME.SparseTensor(coordinates=coords, feats=features)
        return output


class ExpandAs(nn.Module):
    """
    Given a sparse tensor with one dimensional features, expand the
    feature map to given shape and return a newly constructed
    ME.SparseTensor.

        - x (ME.SparseTensor): with x.F.shape[1] == 1
        - shape (tuple)
    """

    def __init__(self):
        super(ExpandAs, self).__init__()

    def forward(self, x, shape):
        device = x.F.device
        features = x.F.expand(*shape)
        output = ME.SparseTensor(
            feats=features, coords_key=x.coords_key, coords_manager=x.coords_man
        )
        return output


class SPICE(torch.nn.Module):

    def __init__(self, cfg, name="spice"):
        super(SPICE, self).__init__()
        setup_cnn_configuration(self, cfg, name)

        self.model_config = cfg.get(name, {})
        self.encoder = UResNetEncoder(self.model_config, name="uresnet_encoder")
        self.embedding_decoder = UResNetDecoder(
            self.model_config, name="embedding_decoder"
        )
        self.seed_decoder = UResNetDecoder(self.model_config, name="seediness_decoder")

        self.skip_classes = self.model_config.get("skip_classes", [2, 3, 4])
        self.num_filters = self.model_config.get("num_filters", 16)
        self.seedDim = self.model_config.get("seediness_dim", 1)
        self.sigmaDim = self.model_config.get("sigma_dim", 1)
        self.seed_freeze = self.model_config.get("seed_freeze", False)
        self.coordConv = self.model_config.get("coordConv", True)

        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()

        self.outputEmbeddings = nn.Sequential(
            ME.MinkowskiBatchNorm(self.num_filters, **self.norm_args),
            ME.MinkowskiLinear(self.num_filters, self.D + self.sigmaDim, bias=False),
        )

        self.outputSeediness = nn.Sequential(
            ME.MinkowskiBatchNorm(self.num_filters, **self.norm_args),
            ME.MinkowskiLinear(self.num_filters, self.seedDim, bias=False),
        )

        if self.seed_freeze:
            print("Seediness Branch Freezed")
            for p in self.seed_decoder.parameters():
                p.requires_grad = False
            for p in self.outputSeediness.parameters():
                p.requires_grad = False

    def filter_class(self, input):
        """
        Filter classes according to segmentation label.
        """
        point_cloud, label = input
        mask = torch.ones(len(label), dtype=bool, device=point_cloud.device)
        for c in self.skip_classes:
            mask &= label[:, -1] != c

        return [point_cloud[mask], label[mask]]

    def forward(self, input):

        point_cloud, _ = self.filter_class(input)
        device = point_cloud.device

        coords = point_cloud[:, 0 : self.D + 1].to(device).int()
        features = point_cloud[:, self.D + 1 :].view(-1, 1)

        normalized_coords = (
            coords[:, 1 : self.D + 1] - float(self.spatial_size) / 2
        ) / (float(self.spatial_size) / 2)
        normalized_coords = normalized_coords
        if self.coordConv:
            features = torch.cat([normalized_coords, features], dim=1)

        x = ME.SparseTensor(features, coordinates=coords)

        encoderOutput = self.encoder(x)
        encoderTensors = encoderOutput["encoderTensors"]
        finalTensor = encoderOutput["finalTensor"]
        features_cluster = self.embedding_decoder(finalTensor, encoderTensors)
        features_seediness = self.seed_decoder(finalTensor, encoderTensors)

        embeddings = self.outputEmbeddings(features_cluster[-1])
        embeddings_feats = embeddings.F
        embeddings_feats[:, : self.D] = self.tanh(embeddings_feats[:, : self.D])
        embeddings_feats[:, : self.D] += normalized_coords
        seediness = self.outputSeediness(features_seediness[-1])

        res = {
            "embeddings": [embeddings_feats[:, : self.D]],
            "seediness": [self.sigmoid(seediness.F)],
            "margins": [2 * self.sigmoid(embeddings_feats[:, self.D :])],
        }
        return res



"""Draw detectors based on their geometry definition."""

import time
from typing import List, Optional, Union

import numpy as np
import plotly.graph_objs as go
from plotly.basedatatypes import BaseTraceType

from spine.data import Meta
from spine.geo import GeoManager, Geometry
from spine.vis.cylinder import cylinder_traces

from .box import box_traces
from .ellipsoid import ellipsoid_traces
from .layout import layout3d

__all__ = ["GeoDrawer"]


class GeoDrawer:
    """Handles drawing all things related to the detector geometry.

    This class is loads a :class:`Geometry` object once from a geometry file
    and uses it to represent all things related to the detector geometry:
    - TPC boundaries
    - Optical detectors
    - CRT detectors

    Attributes
    ----------
    geo : Geometry
        The underlying detector geometry
    detector_coords : bool
        Whether or not to use detector coordinates (True) or pixel indices
    """

    def __init__(self, geo: Optional[Geometry] = None, detector_coords: bool = True):
        """Initializes the underlying detector :class:`Geometry` object.

        Parameters
        ----------
        geo : Geometry, optional
            If provided, this Geometry instance is used.
            If None, the global GeoManager instance is used.
        detector_coords : bool, default False
            If False, the coordinates are converted to pixel indices
        """
        # Fetch the geometry instance, if need be
        if geo is None:
            self.geo = GeoManager.get_instance()
        else:
            assert isinstance(
                geo, Geometry
            ), "The `geo` parameter must be a Geometry instance."
            self.geo = geo

        # Store whether to use detector coordinates or not
        self.detector_coords = detector_coords

    def show(
        self,
        meta: Optional[Meta] = None,
        tpc: bool = True,
        optical: bool = True,
        crt: bool = True,
        **kwargs,
    ):
        """Displays the detector geometry in a 3D plotly figure.

        Parameters
        ----------
        meta : Meta, optional
            Metadata information (only needed if pixel_coordinates is True)
        tpc : bool, default True
            Whether or not to include TPC traces
        optical : bool, default True
            Whether or not to include optical detector traces
        crt : bool, default True
            Whether or not to include CRT detector traces
        **kwargs : dict, optional
            Additional arguments to pass to layout3d
        """

        # Get all the detector traces
        traces = self.traces(
            meta=meta,
            tpc=tpc,
            optical=optical,
            crt=crt,
        )

        # Initialize the layout
        layout = layout3d(
            geo=self.geo,
            use_geo=True,
            meta=meta,
            detector_coords=self.detector_coords,
            show_optical=optical and self.geo.optical is not None,
            show_crt=crt and self.geo.crt is not None,
            **kwargs,
        )

        # Build the 3D layout
        fig = go.Figure(data=traces, layout=layout)

        # Show the figure
        fig.show()

    def traces(
        self,
        meta: Optional[Meta] = None,
        tpc: bool = True,
        optical: bool = True,
        crt: bool = True,
    ) -> List[BaseTraceType]:
        """Returns all traces associated with the detector geometry.

        Parameters
        ----------
        meta : Meta, optional
            Metadata information (only needed if pixel_coordinates is True)
        tpc : bool, default True
            Whether or not to include TPC traces
        optical : bool, default True
            Whether or not to include optical detector traces
        crt : bool, default True
            Whether or not to include CRT detector traces

        Returns
        -------
        List[BaseTraceType]
            List of detector traces
        """
        traces = []
        if tpc:
            traces += self.tpc_traces(meta=meta)
        if optical and self.geo.optical is not None:
            traces += self.optical_traces(meta=meta)
        if crt and self.geo.crt is not None:
            traces += self.crt_traces(meta=meta)

        return traces

    def tpc_traces(
        self,
        meta: Optional[Meta] = None,
        draw_faces: bool = False,
        shared_legend: bool = True,
        name: str = "TPC",
        color: Union[int, str, np.ndarray] = "rgba(0,0,0,0.150)",
        linewidth: int = 5,
        **kwargs,
    ) -> List[BaseTraceType]:
        """Function which produces a list of traces which represent the TPCs in
        a 3D event display.

        Parameters
        ----------
        meta : Meta, optional
            Metadata information (only needed if pixel_coordinates is True)
        draw_faces : bool, default False
            Weather or not to draw the box faces, or only the edges
        shared_legend : bool, default True
            If True, the legend entry in plotly is shared between all the
            TPC volumes
        name : str, default 'TPC'
            Name of the TPC volumes
        color : Union[int, str, np.ndarray]
            Color of boxes or list of color of boxes
        linewidth : int, default 2
            Width of the box edge lines
        **kwargs : dict, optional
            List of additional arguments to pass to
            spine.viusalization.boxes.box_traces

        Returns
        -------
        List[BaseTraceType]
            List of detector traces (one per TPC)
        """
        # Load the list of TPC boundaries
        boundaries = np.stack([c.boundaries for c in self.geo.tpc.chambers])

        # If required, convert to pixel coordinates
        if not self.detector_coords:
            assert meta is not None, (
                "Must provide meta information to convert the TPC "
                "boundaries to pixel coordinates."
            )
            boundaries = meta.to_px(boundaries.transpose(0, 2, 1)).transpose(0, 2, 1)

        # Get a trace per TPC volume
        detectors = box_traces(
            boundaries[..., 0],
            boundaries[..., 1],
            draw_faces=draw_faces,
            color=color,
            linewidth=linewidth,
            shared_legend=shared_legend,
            name=name,
            **kwargs,
        )

        return detectors

    def optical_traces(
        self,
        meta: Optional[Meta] = None,
        shared_legend: bool = True,
        legendgroup: Optional[str] = None,
        name: str = "Optical",
        color: Union[int, str, np.ndarray] = "rgba(0,0,255,0.25)",
        hovertext: Optional[Union[str, List[str]]] = None,
        cmin: Optional[float] = None,
        cmax: Optional[float] = None,
        zero_supress: bool = False,
        volume_id: Optional[int] = None,
        **kwargs,
    ) -> List[BaseTraceType]:
        """Function which produces a list of traces which represent the optical
        detectors in a 3D event display.

        Parameters
        ----------
        meta : Meta, optional
            Metadata information (only needed if pixel_coordinates is True)
        shared_legend : bool, default True
            If True, the legend entry in plotly is shared between all the
            optical volumes
        legendgroup : str, optional
            Legend group to be shared between all boxes
        name : str, default 'Optical'
            Name of the optical volumes
        color : Union[int, str, np.ndarray]
            Color of optical detectors or list of color of optical detectors
        hovertext : Union[str, List[str]], optional
            Label or list of labels associated with each optical detector
        cmin : float, optional
            Minimum value along the color scale
        cmax : float, optional
            Maximum value along the color scale
        zero_supress : bool, default False
            If `True`, do not draw optical detectors that are not activated
        volume_id : int, optional
            Specifies which optical volume to represent. If not specified, all
            the optical volumes are drawn
        **kwargs : dict, optional
            List of additional arguments to pass to
            spine.vis.ellipsoid.ellipsoid_traces or spine.vis.box.box_traces

        Returns
        -------
        List[plotly.graph_objs.Mesh3D]
            List of optical detector traces (one per optical detector)
        """
        # Check that there is optical detectors to draw
        assert (
            self.geo.optical is not None
        ), "This geometry does not have optical detectors to draw."

        # Fetch the optical element positions and sizes
        if volume_id is None:
            positions = self.geo.optical.positions
        else:
            positions = self.geo.optical.volumes[volume_id].positions
        half_sizes = self.geo.optical.sizes / 2

        # If there is more than one detector shape, fetch shape IDs
        shape_ids = self.geo.optical.shape_ids

        # Convert the positions to pixel coordinates, if needed
        if not self.detector_coords:
            assert meta is not None, (
                "Must provide meta information to convert the optical "
                "element positions/sizes to pixel coordinates."
            )
            positions = meta.to_px(positions)
            half_sizes = half_sizes / meta.size

        # Check that the colors provided fix the appropriate range
        if color is not None and isinstance(color, (list, tuple, np.ndarray)):
            assert len(color) == len(
                positions
            ), "Must provide one value for each optical detector."

        # Build the hovertext vectors
        if hovertext is not None:
            if not isinstance(hovertext, (list, tuple, np.ndarray)):
                hovertext = [hovertext] * len(positions)
            elif len(hovertext) != len(positions):
                raise ValueError(
                    "The `hovertext` attribute should be provided as a scalar, "
                    "one value per point or one value per optical detector."
                )

        else:
            hovertext = [f"PD ID: {i}" for i in range(len(positions))]
            if isinstance(color, (list, tuple, np.ndarray)):
                for i, hc in enumerate(hovertext):
                    hovertext[i] = hc + f"<br>Value: {color[i]:.3f}"

        # If cmin/cmax are not provided, must build them so that all optical
        # detectors share the same colorscale range (not guaranteed otherwise)
        if isinstance(color, (list, tuple, np.ndarray)) and len(color) > 0:
            if cmin is None:
                cmin = np.min(color)
            if cmax is None:
                cmax = np.max(color)

        # If the legend is to be shared, make sure there is a common legend group
        if shared_legend and legendgroup is None:
            legendgroup = "group_" + str(time.time())

        # Draw each of the optical detectors
        traces = []
        for i, shape in enumerate(self.geo.optical.shape):
            # Restrict the positions to those of this shape, if needed
            if shape_ids is None:
                pos = positions
                col = color
                ht = hovertext
            else:
                index = np.where(np.asarray(shape_ids) == i)[0]
                pos = positions[index]
                if isinstance(color, (list, tuple, np.ndarray)):
                    col = color[index]
                else:
                    col = color
                ht = [hovertext[i] for i in index]

            # If zero-supression is requested, only draw the optical detectors
            # which record a non-zero signal
            if zero_supress and isinstance(col, (list, tuple, np.ndarray)):
                index = np.where(np.asarray(col) != 0)[0]
                pos = pos[index]
                col = col[index]
                ht = [ht[i] for i in index]

            # Dispatch the drawing based on the type of optical detector
            hd = half_sizes[i]
            if shape == "box":
                # Convert the positions/sizes to box lower/upper bounds
                lower, upper = pos - hd, pos + hd

                # Build boxes
                traces += box_traces(
                    lower,
                    upper,
                    shared_legend=shared_legend,
                    name=name,
                    color=col,
                    cmin=cmin,
                    cmax=cmax,
                    draw_faces=True,
                    hovertext=ht,
                    legendgroup=legendgroup,
                    **kwargs,
                )

            elif shape == "ellipsoid":
                # Convert the optical detector sizes to a covariance matrix
                covmat = np.diag(hd**2)

                # Build ellipsoids
                traces += ellipsoid_traces(
                    pos,
                    covmat,
                    shared_legend=shared_legend,
                    name=name,
                    color=col,
                    cmin=cmin,
                    cmax=cmax,
                    hovertext=ht,
                    legendgroup=legendgroup,
                    **kwargs,
                )

            elif shape == "disk":
                # Build disks as very flat cylinders
                axis = np.zeros(3, dtype=hd.dtype)
                axis[np.argmin(hd)] = 1.0
                length = 2.0 * hd[np.argmin(hd)]
                diameter = 2.0 * hd[np.argmax(hd)]

                # Build disks
                traces += cylinder_traces(
                    pos,
                    axis,
                    length,
                    diameter,
                    shared_legend=shared_legend,
                    name=name,
                    color=col,
                    cmin=cmin,
                    cmax=cmax,
                    hovertext=ht,
                    legendgroup=legendgroup,
                    **kwargs,
                )

            else:
                raise ValueError(
                    f"Optical detector shape '{shape}' not recognized. "
                    "Should be one of 'box', 'ellipsoid' or 'disk'."
                )

        # Set the legend display options
        if shape_ids is not None:
            # If the legend is shared, ensure that only the first trace shows the legend
            if shared_legend:
                for i, trace in enumerate(traces):
                    if i == 0:
                        trace.showlegend = True
                    else:
                        trace.showlegend = False
            else:
                # If the legend is not shared, ensure that the names are unique
                for i, trace in enumerate(traces):
                    trace.name = f"{name} {i}"

        return traces

    def crt_traces(
        self,
        meta: Optional[Meta] = None,
        draw_faces: bool = True,
        shared_legend: bool = True,
        name: str = "CRT",
        color: Union[int, str, np.ndarray] = "rgba(0,256,256,0.150)",
        draw_ids: Optional[List[int]] = None,
        **kwargs,
    ) -> List[BaseTraceType]:
        """Function which produces a list of traces which represent the optical
        detectors in a 3D event display.

        Parameters
        ----------
        meta : Meta, optional
            Metadata information (only needed if pixel_coordinates is True)
        draw_faces : bool, default True
            Weather or not to draw the box faces, or only the edges
        shared_legend : bool, default True
            If True, the legend entry in plotly is shared between all the
            CRT volumes
        name : str, default 'CRT'
            Name of the CRT volumes
        color : Union[int, str, np.ndarray]
            Color of CRT detectors or list of color of CRT detectors
        draw_ids : List[int], optional
            If specified, only the requested CRT planes are drawn
        **kwargs : dict, optional
            List of additional arguments to pass to
            spine.vis.ellipsoid.ellipsoid_traces or spine.vis.box.box_traces

        Returns
        -------
        List[plotly.graph_objs.Mesh3D]
            List of CRT detector traces (one per CRT element)
        """
        # Check that there are CRT planes to draw
        assert (
            self.geo.crt is not None
        ), "This geometry does not have CRT planes to draw."

        # Load the list of CRT plane boundaries
        boundaries = np.stack([p.boundaries for p in self.geo.crt.planes])

        # If required, convert to pixel coordinates
        if not self.detector_coords:
            assert meta is not None, (
                "Must provide meta information to convert the CRT plane "
                "boundaries to pixel coordinates."
            )
            boundaries = meta.to_px(boundaries.transpose(0, 2, 1)).transpose(0, 2, 1)

        # Restrict the list of boundaries, if requested
        if draw_ids is not None:
            tmp = np.empty(
                (len(draw_ids), *boundaries.shape[1:]), dtype=boundaries.dtype
            )
            for i, idx in enumerate(draw_ids):
                tmp[i] = boundaries[idx]

            boundaries = tmp

        # Get a trace per CRT plane
        detectors = box_traces(
            boundaries[..., 0],
            boundaries[..., 1],
            draw_faces=draw_faces,
            color=color,
            shared_legend=shared_legend,
            name=name,
            **kwargs,
        )

        return detectors



"""Particle truth information post-processors.

They add/correct information stored in LArCV particles.
"""

from warnings import warn

import numpy as np

from .globals import DELTA_SHP, INVAL_ID, INVAL_IDX, INVAL_TID, MICHL_SHP, PDG_TO_PID


def process_particles(
    particles, particle_event, particle_mpv_event=None, neutrino_event=None
):
    """Process Particle object list to add/correct attributes in place.

    Does the following:
    - Checks that the group ID and parent ID are valid, assign -1 if not
    - Checks that the ancestor track ID is valid, assign -1 if not
    - Adds interaction ID information if it is not provided
    - Adds the true neutrino ID this particle came from
    - Adds a simplified enumerated particle species ID
    - Adds a flag as to whether a particle is a primary within its interaction
    - Adds a flag as to whether a particle is a primary within its group

    Notes
    -----
    The following issues have been observed in the wild:
    - In the case of generic (v04) files
        - The parent ID and the interaction ID might not be filled properly (INVAL_IDX)
        - The ancestor track ID might not filled properly (INVAL_TID)
    - In the case of ICARUS/SBND files
        - The parent ID and the interaction ID might not be filled properly (INVAL_ID)
        - The ancestor track ID might not filled properly (INVAL_TID)
    - In the case of DUNE-ND/2x2 files
        - The group ID might not be filled properly (INVAL_ID)

    Parameters
    ----------
    particles : List[Particle]
        (P) List of true particle instances
    particle_event : larcv.EventParticle
        (P) List of true particle instances
    particle_mpv_event : larcv.EventParticle, optional
        (M) List of true MPV particle instances
    neutrino_event : larcv.EventNeutrino, optional
        (N) List of true neutrino instances
    """
    # If the list is empty, there is nothing to do
    if len(particles) == 0:
        return

    # Get the additional attributes
    (
        parent_ids,
        group_ids,
        ancestor_ids,
        interaction_ids,
        nu_ids,
        group_primary_ids,
        inter_primary_ids,
        pids,
    ) = process_particle_event(particle_event, particle_mpv_event, neutrino_event)

    # Update the particles objects in place
    for i, p in enumerate(particles):
        if p.id > -1:
            p.parent_id = parent_ids[i]
            p.group_id = group_ids[i]
            p.ancestor_track_id = ancestor_ids[i]
            p.interaction_id = interaction_ids[i]
            p.nu_id = nu_ids[i]
            p.group_primary = group_primary_ids[i]
            p.interaction_primary = inter_primary_ids[i]
            p.pid = pids[i]


def process_particle_event(
    particle_event, particle_mpv_event=None, neutrino_event=None
):
    """Corrects/fetches attributes for a larcv.EventParticle object.

    Does the following:
    - Checks that the group ID and parent ID are valid, assign -1 if not
    - Checks that the ancestor track ID is valid, assign -1 if not
    - Builds the interaction ID information if it is not provided
    - Gets the true neutrino ID this particle came from
    - Gets a simplified enumerated particle species ID
    - Gets a flag as to whether a particle is a primary within its interaction
    - Gets a flag as to whether a particle is a primary within its group

    Notes
    -----
    The following issues have been observed in the wild:
    - In the case of generic (v04) files
        - The parent ID and the interaction ID might not be filled properly (INVAL_IDX)
        - The ancestor track ID might not filled properly (INVAL_TID)
    - In the case of ICARUS/SBND files
        - The parent ID and the interaction ID might not be filled properly (INVAL_ID)
        - The ancestor track ID might not filled properly (INVAL_TID)
    - In the case of DUNE-ND/2x2 files
        - The group ID might not be filled properly (INVAL_ID)

    Parameters
    ----------
    particle_event : larcv.EventParticle
        (P) List of true particle instances
    particle_mpv_event : larcv.EventParticle, optional
        (M) List of true MPV particle instances
    neutrino_event : larcv.EventNeutrino, optional
        (N) List of true neutrino instances

    Returns
    -------
    group_ids : np.ndarray
        (P) List of group IDs, one per true particle instance
    parent_ids : np.ndarray
        (P) List of parent IDs, one per true particle instance
    ancestor_ids : np.ndarray
        (P) List of ancestor IDs, one per true particle instance
    interaction_ids : np.ndarray
        (P) List of interaction IDs, one per true particle instance
    nu_ids : np.ndarray
        (P) List of neutrino IDs, one per true particle instance
    group_primary_ids : np.ndarray
        (P) List of particle group primary IDs, one per true particle instance
    inter_primary_ids : np.ndarray
        (P) List of particle primary IDs, one per true particle instance
    pids : np.ndarray
        (P) List of particle IDs, one per true particle instance
    """
    # Converts the input to simple python lists of objects
    particles = list(particle_event.as_vector())
    particles_mpv, neutrinos = None, None
    if particle_mpv_event is not None:
        particles_mpv = list(particle_mpv_event.as_vector())
    if neutrino_event is not None:
        neutrinos = list(neutrino_event.as_vector())

    # Check on the parent ID of each particle
    parent_ids = np.array([p.parent_id() for p in particles], dtype=int)
    if np.any(parent_ids == INVAL_ID):
        # This takes care of ICARUS/SBND file issues
        parent_ids[parent_ids == INVAL_ID] = -1
    elif len(particles) <= INVAL_IDX:
        # This takes care of generic file issues
        parent_ids[parent_ids == INVAL_IDX] = -1

    # Check on the group ID of each particle
    # This takes care of DUNE-ND/2x2 file issues
    group_ids = np.array([p.group_id() for p in particles], dtype=int)
    group_ids[group_ids == INVAL_ID] = -1

    # Check on the ancestor track ID of each particle
    # This takes care of generic/ICARUS/SBND file issues
    ancestor_ids = np.array([p.ancestor_track_id() for p in particles], dtype=int)
    ancestor_ids[ancestor_ids == INVAL_TID] = -1

    # Get the mask of valid particle labels
    valid_mask = get_valid_mask(particles)

    # Get the interaction ID of each particle
    interaction_ids = get_interaction_ids(particles, valid_mask)

    # Get the neutrino ID of each particle
    nu_ids = get_nu_ids(particles, interaction_ids, particles_mpv, neutrinos)

    # Get the group primary status of each particle
    group_primary_ids = get_group_primary_ids(particles, valid_mask)

    # Get the interaction primary status of each particle
    inter_primary_ids = get_inter_primary_ids(particles, valid_mask)

    # Get the particle species (PID) of each particle
    pids = get_particle_ids(particles, valid_mask)

    # Return
    return (
        parent_ids,
        group_ids,
        ancestor_ids,
        interaction_ids,
        nu_ids,
        group_primary_ids,
        inter_primary_ids,
        pids,
    )


def get_valid_mask(particles):
    """Gets a mask corresponding to particles with valid labels.

    This function checks that the particle labels have been filled properly at
    the Supera level. It checks that the ancestor track ID of each particle is
    not an invalid number and that the ancestor creation process is filled.

    Parameters
    ----------
    particles : List[larcv.Particle]
        (P) List of true particle instances

    Results
    -------
    np.ndarray
        (P) Boolean list of validity, one per true particle instance
    """
    # If there are no particles, nothing to do here
    if len(particles) == 0:
        return np.empty(0, dtype=bool)

    # If the interaction IDs are set in the particle tree, simply use that
    interaction_ids = np.array([p.interaction_id() for p in particles], dtype=int)
    if np.any((interaction_ids != INVAL_ID) & (interaction_ids != INVAL_IDX)):
        return interaction_ids != INVAL_ID

    # Otherwise, check that the ancestor track ID and creation process are valid
    mask = np.array([p.ancestor_track_id() != INVAL_TID for p in particles])
    mask &= np.array([bool(p.ancestor_creation_process()) for p in particles])

    return mask


def get_interaction_ids(particles, valid_mask=None):
    """Gets the interaction ID of each particle.

    If the `interaction_id` attribute of the Particle class is filled,
    it simply uses that quantity.

    Otherwise, it leverages shared ancestor position as a basis for
    interaction building and sets the interaction ID to -1 for particles with
    invalid ancestor track IDs.

    Parameters
    ----------
    particles : List[larcv.Particle]
        (P) List of true particle instances
    valid_mask : np.ndarray, optional
        (P) Particle label validity mask

    Results
    -------
    np.ndarray
        (P) List of interaction IDs, one per true particle instance
    """
    # If there are no particles, nothing to do here
    if len(particles) == 0:
        return np.empty(0, dtype=int)

    # Compute the validity mask if it is not provided
    if valid_mask is None:
        valid_mask = get_valid_mask(particles)

    # If the interaction IDs are set in the particle tree, simply use that
    interaction_ids = np.array([p.interaction_id() for p in particles], dtype=int)
    if np.any((interaction_ids != INVAL_ID) & (interaction_ids != INVAL_IDX)):
        interaction_ids[~valid_mask] = -1
        return interaction_ids

    # Otherwise, define interaction IDs on the basis of sharing
    # an ancestor vertex position
    anc_pos = np.vstack([get_coords(p.ancestor_position()) for p in particles])
    interaction_ids = np.unique(anc_pos, axis=0, return_inverse=True)[-1]

    # Now set the interaction ID of particles with an undefined ancestor to -1
    interaction_ids[~valid_mask] = -1

    return interaction_ids


def get_nu_ids(particles, interaction_ids, particles_mpv=None, neutrinos=None):
    """Gets the neutrino-like ID of each partcile

    Convention: -1 for non-neutrinos, neutrino index for others

    If a list of multi-particle vertex (MPV) particles or neutrinos is
    provided, that information is leveraged to identify which interactions
    are neutrino-like and which are not.

    If `particles_mpv` and `neutrinos` are not specified, it assumes that
    only neutrino-like interactions have more than one true primary
    particle in a single interaction.

    Parameters
    ----------
    particles : List[larcv.Particle]
        (P) List of true particle instances
    interaction_ids : np.ndarray
        (P) Array of interaction ID values, one per true particle instance
    particles_mpv : List[larcv.Particle], optional
        (M) List of true MPV particle instances
    neutrinos : list(larcv.Neutrino), optional
        (N) List of true neutrino instances

    Results
    -------
    np.ndarray
        (P) List of neutrino IDs, one per true particle instance
    """
    # If there are no particles, nothing to do here
    if len(particles) == 0:
        return np.empty(0, dtype=int)

    # Make sure there is only either MPV particles or neutrinos specified, not both
    assert particles_mpv is None or neutrinos is None, (
        "Do not specify both particle_mpv_event and neutrino_event "
        "in `get_neutrino_ids`. Can only use one or the other."
    )

    # Initialize neutrino IDs
    nu_ids = -np.ones(len(particles), dtype=int)
    if particles_mpv is None and neutrinos is None:
        # Warn that this is ad-hoc
        warn(
            "Neutrino IDs are being produced on the basis of interaction "
            "multiplicity (i.e. neutrino if >= 2 primaries). This is "
            "not an exact method and might lead to unexpected results."
        )

        # Loop over the interactions
        primary_ids = get_inter_primary_ids(particles, interaction_ids > -1)
        nu_id = 0
        for i in np.unique(interaction_ids):
            # If the interaction ID is invalid, skip
            if i < 0:
                continue

            # If there are at least two primaries, the interaction is nu-like
            inter_index = np.where(interaction_ids == i)[0]
            if np.sum(primary_ids[inter_index] == 1) > 1:
                nu_ids[inter_index] = nu_id
                nu_id += 1

    else:
        # Find the reference positions to gauge if a particle comes from a
        # nu-like interaction
        ref_ids, ref_pos = None, None
        if particles_mpv and len(particles_mpv) > 0:
            ref_pos = np.vstack([get_coords(p.position()) for p in particles_mpv])
            ref_pos = np.unique(ref_pos, axis=0)

        elif neutrinos and len(neutrinos) > 0:
            if (
                hasattr(neutrinos[0], "interaction_id")
                and neutrinos[0].interaction_id() != INVAL_IDX
            ):
                ref_ids = np.array([n.interaction_id() for n in neutrinos])
            else:
                ref_pos = np.vstack([get_coords(n.position()) for n in neutrinos])

        # If an interaction ID is provided for neutrinos, the matching is trivial
        if ref_ids is not None:
            for i in np.unique(interaction_ids):
                # If the interaction is invalid, skip
                if i < 0:
                    continue

                # Loop over positions in the interaction find a reference match
                inter_index = np.where(interaction_ids == i)[0]
                for nu_id, ref_id in enumerate(ref_ids):
                    if i == ref_id:
                        nu_ids[inter_index] = nu_id
                        break

        # Otherwise, if a particle in an interaction shares its ancestor
        # position with an MPV particle or a neutrino, the whole interaction
        # is labeled as a nu-like interaction.
        if ref_pos is not None:
            warn(
                "Neutrino IDs are being produced on the basis of floating "
                "point agreement between particle ancestor positions and "
                "neutrino positions. For safety, should provide a matching "
                "interaction_id to the larcv.Neutrino object."
            )

            anc_pos = np.vstack([get_coords(p.ancestor_position()) for p in particles])
            for i in np.unique(interaction_ids):
                # If the interaction is invalid, skip
                if i < 0:
                    continue

                # Loop over positions in the interaction find a reference match
                inter_index = np.where(interaction_ids == i)[0]
                for ref_id, pos in enumerate(ref_pos):
                    if np.any((anc_pos[inter_index] == pos).all(axis=1)):
                        nu_ids[inter_index] = ref_id
                        break

    return nu_ids


def get_group_primary_ids(particles, valid_mask=None):
    """Gets the group primary status of particle fragments.

    This could be handled somewhere else (e.g. Supera).

    Parameters
    ----------
    particles : List[larcv.Particle]
        (P) List of true particle instances
    valid_mask : np.ndarray, optional
        (P) Particle label validity mask

    Results
    -------
    np.ndarray
        (P) List of particle group primary IDs, one per true particle instance
    """
    # Compute the validity mask if it is not provided
    if valid_mask is None:
        valid_mask = get_valid_mask(particles)

    # Loop over the list of particle groups
    primary_ids = np.zeros(len(particles), dtype=int)
    group_ids = np.array([p.group_id() for p in particles], dtype=int)
    for group_id in np.unique(group_ids):
        # Check that the group ID is within the expected range
        group_index = np.where(group_ids == group_id)[0]
        if group_id != INVAL_ID and group_id > len(particles) - 1:
            warn(
                f"Bad group ID ({group_id}) not matching INVAL_ID "
                f"({INVAL_ID}). This may happen for old files."
            )
            primary_ids[group_index] = -1
            continue

        # If the particle group has invalid labeling, the concept of group
        # primary is ill-defined
        if group_id == INVAL_ID or not valid_mask[group_id]:
            primary_ids[group_index] = -1
            continue

        # If a group originates from a Delta or a Michel, it has a primary
        group_p = particles[group_id]
        if group_p.shape() == MICHL_SHP or group_p.shape() == DELTA_SHP:
            primary_ids[group_id] = 1
            continue

        # If a particle group's parent fragment is the first in time,
        # it is a valid primary. TODO: use first step time.
        clust_times = np.array([particles[i].t() for i in group_index])
        min_id = np.argmin(clust_times)
        if group_index[min_id] == group_id:
            primary_ids[group_id] = 1

    return primary_ids


def get_inter_primary_ids(particles, valid_mask=None):
    """Gets the interaction primary ID for each particle.

    Parameters
    ----------
    particles : List[larcv.Particle]
        (P) List of true particle instances
    valid_mask : np.ndarray, optional
        (P) Particle label validity mask

    Results
    -------
    np.ndarray
        (P) List of particle primary IDs, one per true particle instance
    """
    # Compute the validity mask if it is not provided
    if valid_mask is None:
        valid_mask = get_valid_mask(particles)

    # Loop over the list of particles
    primary_ids = -np.ones(len(particles), dtype=int)
    for i, p in enumerate(particles):
        # If the particle has invalid labeling, it has invalid primary status
        group_id = p.group_id()
        if group_id == INVAL_ID or not valid_mask[i]:
            continue

        # Check that the group ID is within the expected range
        if group_id > len(particles) - 1:
            warn(
                f"Bad group ID ({group_id}) not matching INVAL_ID "
                f"({INVAL_ID}). This may happen for old files."
            )
            continue

        # Tag particles originating from a very short-lived particle as primary
        group_p = particles[group_id]
        if (
            abs(group_p.parent_pdg_code()) in [111, 113, 221, 331, 3212]
            and group_p.parent_track_id() == group_p.ancestor_track_id()
        ):
            primary_ids[i] = 1
            continue

        # If the origin of a particle agrees with the origin of its ancestor,
        # label as primary
        primary_ids[i] = p.track_id() == p.ancestor_track_id()

    return primary_ids


def get_particle_ids(particles, valid_mask=None):
    """Gets a particle species ID (PID) for each particle.

    This function ensures:
    - All shower daughters are labeled the same as their primary. This
      makes sense as otherwise an electron primary gets overruled by
      its many photon daughters (voxel-wise majority vote). This can
      lead to problems as, if an electron daughter is not clustered with
      the primary, it is labeled electron, which is counter-intuitive.
      This is handled downstream with the high_purity flag.
    - Particles that are not in the list target are labeled -1

    Parameters
    ----------
    particles : List[Particle]
        (P) List of true particle instances
    valid_mask : np.ndarray, optional
        (P) Particle label validity mask

    Returns
    -------
    np.ndarray
        (P) List of particle IDs, one per true particle instance
    """
    # Compute the validity mask if it is not provided
    if valid_mask is None:
        valid_mask = get_valid_mask(particles)

    # Loop over the list of particles
    particle_ids = -np.ones(len(particles), dtype=int)
    for i, p in enumerate(particles):
        # If the primary ID is invalid, skip
        group_id = p.group_id()
        if group_id == INVAL_ID or not valid_mask[i]:
            continue

        # Check that the group ID is within the expected range
        if group_id > len(particles) - 1:
            warn(
                f"Bad group ID ({group_id}) not matching INVAL_ID "
                f"({INVAL_ID}). This may happen for old files."
            )
            continue

        # If the particle type exists in the predefined list, assign
        t = particles[group_id].pdg_code()
        if t in PDG_TO_PID.keys():
            particle_ids[i] = PDG_TO_PID[t]

    return particle_ids


def get_coords(position):
    """Gets the coordinates of a larcv.Vertex object.

    Parameters
    ----------
    position : larcv.Vertex
        Encodes the position of a point with attributes x, y, z and t

    Returns
    -------
    List[float]
        Coordinates of the point (x, y, z)
    """
    return np.array([getattr(position, a)() for a in ["x", "y", "z"]])



"""Module with losses which are not generically provided by `PyTorch`."""

import torch
from torch import nn

from spine.utils.weighting import get_class_weights

from .act_norm import act_factory

__all__ = [
    "LogRMSE",
    "BerHuLoss",
    "BinaryDiceLoss",
    "BinaryLogDiceLoss",
    "BinaryMincutLoss",
    "BinaryLogDiceCELoss",
    "BinaryLogDiceCEMincutLoss",
    "FocalLoss",
]


class LogRMSE(nn.modules.loss._Loss):
    """Applies RMSE loss to in the log space for regression tasks."""

    name = "log_rmse"

    def __init__(self, reduction="none", eps=1e-7):
        """Initialize the loss function parameters.

        Parameters
        ----------
        reduction : str, default 'none'
            Reduction function to apply to the output
        eps : float, default 1e-7
            Offset to apply to the predictions/labels before passing them
            through the MSE loss function.
        """
        # Initialize the parent class
        super().__init__()

        # Store the attributes
        self.reduction = reduction
        self.eps = eps

        # Initialize the underlying MSE loss function
        self.mseloss = nn.MSELoss(reduction="none")

    def forward(self, inputs, targets):
        """Pass predictions/labels through the loss function.

        Parameters
        ----------
        inputs : torch.Tensor
            (N) Values predicted by the network
        targets : torch.Tensor
            (N) Regression targets

        Returns
        -------
        torch.Tensor
            Loss value or array of loss values (if no reduction)
        """
        # Move the input/target to the log space
        x = torch.log(inputs + self.eps)
        y = torch.log(targets + self.eps)

        # Compute the RMSE loss
        out = self.mseloss(x, y)
        out = torch.sqrt(out + self.eps)

        # Return the appropriate reduction output
        if self.reduction == "none":
            return out
        elif self.reduction == "mean":
            return out.mean()
        elif self.reduction == "sum":
            return out.sum()
        else:
            raise ValueError("Reduction function not recognized:", self.reduction)


class BerHuLoss(nn.modules.loss._Loss):
    """Applies the BerHu loss."""

    name = "berhu"

    def __init__(self, threshold=0.2, reduction="none"):
        """Initialize the loss function parameters.

        Parameters
        ----------
        threshold : float, default 0.2
            Fraction of the maximum loss value to use as a threshold
        reduction : str, default 'none'
            Reduction function to apply to the output
        """
        # Initialize the parent class
        super().__init__()

        # Store the attributes
        self.threshold = threshold
        self.reduction = reduction

    def forward(self, inputs, targets):
        """Pass predictions/labels through the loss function.

        Parameters
        ----------
        inputs : torch.Tensor
            (N) Values predicted by the network
        targets : torch.Tensor
            (N) Regression targets

        Returns
        -------
        torch.Tensor
            Loss value or array of loss values (if no reduction)
        """
        # Compute the L1 loss
        norm = torch.abs(inputs - targets)

        # If the norm array is of length 0, nothing to do
        if len(norm) == 0:
            return norm.sum()

        # Apply different losses below and above the threshold
        c = norm.max() * self.threshold
        out = torch.where(norm <= c, norm, (norm**2 + c**2) / (2.0 * c))

        # Return the appropriate reduction output
        if self.reduction == "none":
            return out
        elif self.reduction == "mean":
            return out.mean()
        elif self.reduction == "sum":
            return out.sum()
        else:
            raise ValueError("Reduction function not recognized:", self.reduction)


class BinaryDiceLoss(nn.modules.loss._Loss):
    """Applies the binary Dice Loss.

    The Dice loss is derived from the Dice Similarity Coefficient, also known
    as the Sorensen–Dice coefficient, which is a statistical measure used to
    compare the similarity of two samples.
    """

    name = "binary_dice"

    def __init__(self, eps=1e-6, squared_pred=True, activation="sigmoid"):
        """Initialize the loss function parameters.

        Parameters
        ----------
        eps : float, default 1e-6
            Regularization constant for the ratio
        """
        # Initialize the parent class
        super().__init__()

        # Store the loss parameters
        self.eps = eps
        self.squared_pred = squared_pred

        # Intitialize the activation layer
        self.act = act_factory(activation)

    def forward(self, logits, targets):
        """Pass predictions/labels through the loss function.

        Parameters
        ----------
        logits : torch.Tensor
            (N) Values predicted by the network
        targets : torch.Tensor
            (N) Regression targets

        Returns
        -------
        torch.Tensor
            Loss value
        """
        # Compute probability measures using the activation function
        probas = self.act(logits)

        # Compute the dice loss
        inter = (probas * targets).sum()
        if not self.squared_pred:
            den = probas.sum() + targets.sum()
        else:
            den = (probas**2).sum() + (targets**2).sum()

        return 1.0 - (2 * inter + self.eps) / (den + self.eps)


class BinaryLogDiceLoss(BinaryDiceLoss):
    """Applies the binary log Dice loss.

    This class inherits from the standard :class:`BinaryDiceLoss` and simply
    passes it through a logarithm.
    """

    name = "binary_log_dice"

    def forward(self, logits, targets):
        """Pass predictions/labels through the loss function.

        Parameters
        ----------
        logits : torch.Tensor
            (N) Values predicted by the network
        targets : torch.Tensor
            (N) Regression targets

        Returns
        -------
        torch.Tensor
            Loss value
        """
        dice = super().forward(logits, targets)
        dice = torch.clamp(dice, min=self.eps, max=1.0 - self.eps)
        return -torch.log(1.0 - dice)


class BinaryMincutLoss(nn.modules.loss._Loss):
    """Applies the min-cut loss.

    This is a very basic loss of 1. - interesection between the output
    probabilities and the target domain.
    """

    name = "binary_mincut"

    def __init__(self, activation="sigmoid"):
        """Initialize the loss function parameters.

        Parameters
        ----------
        eps : float, default 1e-6
            Regularization constant for the ratio
        """
        # Initialize the parent class
        super().__init__()

        # Intitialize the activation layer
        self.act = act_factory(activation)

    def forward(self, logits, targets):
        """Pass predictions/labels through the loss function.

        Parameters
        ----------
        logits : torch.Tensor
            (N) Values predicted by the network
        targets : torch.Tensor
            (N) Regression targets

        Returns
        -------
        torch.Tensor
            Loss value
        """
        # Compute probability measures using the activation function
        probas = self.act(logits)

        # Compute the mincut loss
        return 1.0 - (probas * targets).sum()


class BinaryLogDiceCELoss(nn.modules.loss._Loss):
    """Applies the binary log Dice loss and the cross-entropy loss.

    This class inherits from the :class:`BinaryLogDiceLoss` and adds
    a cross-entropy loss on top of it, with some configurable weights.
    """

    name = "binary_log_dice_ce"

    def __init__(self, log_dice=None, bce=None, reduction="mean", w_dice=0.8, w_ce=0.2):
        """Initialize the loss function parameters.

        Parameters
        ----------
        log_dice : dict, optional
            Parameters to pass to the :class:`BinaryLogDiceLoss`
        bce : dict, optional
            Parameters to pass to the :class:`torch.nn.BCEWithLogitsLoss`
        reduction : str, default 'mean'
            Reduction function to apply tot he BCE loss
        w_dice : float, default 0.8
            Prefacor to be applied to the log Dice loss
        w_ce : float, default 0.2
            Prefactor to be applied to the binary cross-entropy loss
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the binary log DICE loss
        log_dice = (log_dice is not None) or {}
        self.log_dice = BinaryLogDiceLoss(**log_dice)

        # Initiliaze the binary cross-entropy loss
        bce = (bce is not None) or {}
        self.bce = torch.nn.BCEWithLogitsLoss(**bce, reduction=reduction)

        # Store the loss component weights
        self.w_dice = w_dice
        self.w_ce = w_ce

    def forward(self, logits, targets):
        """Pass predictions/labels through the loss function.

        Parameters
        ----------
        logits : torch.Tensor
            (N) Values predicted by the network
        targets : torch.Tensor
            (N) Regression targets

        Returns
        -------
        torch.Tensor
            Loss value
        """
        # Compute the log dice loss
        log_dice = self.log_dice(logits, targets)

        # Compute the mean binary cross-entropy loss
        bce = self.bce(logits, targets.float())

        # Combine the losses
        return self.w_dice * log_dice + self.w_ce * bce


class BinaryLogDiceCEMincutLoss(nn.modules.loss._Loss):
    """Applies the binary log Dice loss, cross-entropy loss and mincut loss.

    This class inherits from the :class:`BinaryLogDiceCELoss` and adds
    a mincut loss on top of it, with some configurable weights.
    """

    name = "binary_log_dice_ce_mincut"

    def __init__(self, w_mincut=None, **kwargs):
        """Initialize the loss function parameters.

        Parameters
        ----------
        w_mincut : float, default 1.0
            Prefacor to be applied to the Mincut loss
        mincut : dict, optional
            Parameters to pass to the :class:`Mincut`
        **kwargs : dict, optional
            Parameters to pass to the :class:`BinaryLogDiceCELoss`
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Initiliaze the binary Mincut loss
        mincut = (mincut is not None) or {}
        self.mincut = MincutLoss(**mincut)

        # Check that the activation functions are consistent
        assert self.log_dice.act == self.mincut.act, (
            "The log Dice loss and Mincut loss must have the same "
            "activation functions."
        )

        # Store the loss component weights
        self.w_mincut = w_mincut

    def forward(self, logits, targets):
        """Pass predictions/labels through the loss function.

        Parameters
        ----------
        logits : torch.Tensor
            (N) Values predicted by the network
        targets : torch.Tensor
            (N) Regression targets

        Returns
        -------
        torch.Tensor
            Loss value
        """
        # Compute the log dice loss
        log_dice = self.log_dice(logits, targets)

        # Compute the mean binary cross-entropy loss
        bce = self.bce(logits, targets.float())

        # Compute the mincut loss
        mincut = self.mincut(logits, targets)

        # Combine the losses
        return self.w_dice * log_dice + self.w_ce * bce + self.w_mincut * mincut


class BinaryFocalLoss(nn.modules.loss._Loss):
    """Applies the focal loss.

    Original Paper: https://arxiv.org/abs/1708.02002
    """

    name = "binary_focal"

    def __init__(
        self, alpha=1, gamma=2, logits=False, balance_loss=False, reduction=True
    ):
        """Initialize the loss function parameters.

        Parameters
        ----------
        alpha : float, default 1
            Overall loss scaling factor
        gamma : float, default 2
            Overall power to apply to the score prefactor
        logits : bool, default False
            If `True`, the output of the network is considered to be logits
        balance_loss : bool, default False
            If `True`, weights are applied to the loss to account for class imbalance
        reduction : str, default 'none'
            Reduction function to apply to the output
        """
        # Initialize the parent class
        super().__init__()

        # Store the loss parameters
        self.alpha = alpha
        self.gamma = gamma
        self.balance_loss = balance_loss
        self.reduction = reduction

        # Initialize the BCE layer
        if logits:
            self.bce = torch.nn.BCELoss(reduction="none")
        else:
            self.bce = torch.nn.BCEWithLogitsLoss(reduction="none")

    def forward(self, inputs, targets):
        """Pass predictions/labels through the loss function.

        Parameters
        ----------
        inputs : torch.Tensor
            (N) Values predicted by the network
        targets : torch.Tensor
            (N) Regression targets

        Returns
        -------
        torch.Tensor
            Loss value
        """
        # Compute the BCE loss
        bce = self.bce(inputs, targets.float())

        # Compute cross-entropy loss to softmax scores, compute focal lsos
        pt = torch.exp(-bce)
        out = self.alpha * (1 - pt) ** self.gamma * BCE_loss

        # If requested, balance classes
        if self.balance_loss:
            with torch.no_grad():
                weights = get_class_weights(targets, 2, per_class=False)

        # Return the appropriate reduction output
        if self.reduction == "none":
            return out
        elif self.reduction == "mean":
            return out.mean()
        elif self.reduction == "sum":
            return out.sum()
        else:
            raise ValueError("Reduction function not recognized:", self.reduction)



import torch
import torch.nn as nn

from spine.model.layer.cluster import loss_factory
from spine.model.layer.cluster.embeddings import (
    SPICE as SPICE_base,  # TODO why does this live out of this module?
)


class SPICE(SPICE_base):

    MODULES = [
        "network_base",
        "uresnet_encoder",
        "embedding_decoder",
        "seediness_decoder",
    ]

    def __init__(self, cfg):
        super(SPICE, self).__init__(cfg)


class SPICELoss(nn.Module):
    """
    Loss function for Proposal-Free Mask Generators.
    """

    def __init__(self, cfg, name="spice_loss"):
        super(SPICELoss, self).__init__()

        self.model_config = cfg.get("spice", {})
        self.skip_classes = self.model_config.get("skip_classes", [2, 3, 4])

        self.loss_config = cfg.get(name, {})
        self.loss_func_name = self.loss_config.get("name", "se_lovasz_inter")
        self.loss_func = loss_factory(self.loss_func_name)
        self.loss_func = self.loss_func(cfg)
        # print(self.loss_func)

    def class_mask(self, cluster_label):
        """
        Filter classes according to segmentation label.
        """
        mask = torch.ones(len(cluster_label), dtype=bool, device=cluster_label.device)
        for c in self.skip_classes:
            mask &= cluster_label[:, -1] != c

        return mask

    def forward(self, result, cluster_label):
        mask = self.class_mask(cluster_label[0])
        segment_label = [cluster_label[0][mask][:, [0, 1, 2, 3, -1]]]
        group_label = [cluster_label[0][mask][:, [0, 1, 2, 3, 5]]]
        return self.loss_func(result, segment_label, group_label)



"""Analysis script used to store the reconstruction output to CSV files."""

from spine.ana.base import AnaBase
from spine.data.out import (
    RecoFragment,
    RecoInteraction,
    RecoParticle,
    TruthFragment,
    TruthInteraction,
    TruthParticle,
)

__all__ = ["SaveAna"]


class SaveAna(AnaBase):
    """Class which simply saves reconstructed objects (and their matches)."""

    name = "save"

    # Valid match modes
    _match_modes = (None, "reco_to_truth", "truth_to_reco", "both", "all")

    # Default object types when a match is not found
    _default_objs = (
        ("reco_fragments", RecoFragment()),
        ("truth_fragments", TruthFragment()),
        ("reco_particles", RecoParticle()),
        ("truth_particles", TruthParticle()),
        ("reco_interactions", RecoInteraction()),
        ("truth_interactions", TruthInteraction()),
    )

    def __init__(
        self,
        obj_type,
        fragment=None,
        particle=None,
        interaction=None,
        lengths=None,
        run_mode="both",
        match_mode="both",
        **kwargs,
    ):
        """Initialize the CSV logging class.

        If any of the `fragments`, `particles` or `interactions` are specified
        as lists of strings, it will be used to restrict the list of
        object attributes which get stored to CSV.

        Parameters
        ----------
        obj_type : Union[str, List[str]], default ['particle', 'interaction']
            Objects to build files from
        fragment : List[str], optional
            List of fragment attributes to store
        particle : List[str], optional
            List of particle attributes to store
        interaction : List[str], optional
            List of interaction attributes to store
        lengths : Dict[str, int], optional
            Lengths to use for variable-length object attributes
        match_mode : str, default 'both'
            If reconstructed and truth are available, specified which matching
            direction(s) should be saved to the log file.
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__(obj_type=obj_type, run_mode=run_mode, **kwargs)

        # Store the matching mode
        self.match_mode = match_mode
        assert match_mode in self._match_modes, (
            f"Invalid matching mode: {self.match_mode}. Must be one "
            f"of {self._match_modes}."
        )
        assert match_mode is None or run_mode == "both", (
            "When storing matches, you must load both reco and truth "
            f"objects, i.e. set `run_mode` to `True`. Got {run_mode}."
        )

        # Store default objects as a dictionary
        self.default_objs = dict(self._default_objs)

        # Store the list of attributes to store for each object type
        attrs = {
            "fragments": fragment,
            "particles": particle,
            "interactions": interaction,
        }
        if run_mode != "both":
            # If there is only one object type, the keys specified are unique
            self.attrs = {f"{run_mode}_{k}": v for k, v in attrs.items()}

        else:
            # If there are multiple object types, down select to attributes
            # each declination of the object knows, as long as either one does
            self.attrs = {}
            for obj_t, attrs_t in attrs.items():
                # Create a list speicific to each object declination
                leftover = set(attrs_t) if attrs_t is not None else None
                for run_mode in ["reco", "truth"]:
                    key = f"{run_mode}_{obj_t}"
                    if attrs_t is not None:
                        all_keys = self.default_objs[key].as_dict().keys()
                        self.attrs[key] = set(attrs_t) & set(all_keys)
                        leftover -= leftover & self.attrs[key]

                    else:
                        self.attrs[key] = attrs_t

                # Check that there are no leftover keys
                assert leftover is None or len(leftover) == 0, (
                    "The following keys were not found in either the reco "
                    f"or the truth {obj_t} : {leftover}"
                )

        # Store the list of variable-length array lengths
        self.lengths = lengths

        # Add the necessary keys associated with matching, if needed
        keys = {}
        if match_mode is not None:
            for prefix in self.prefixes:
                for obj_name in self.obj_type:
                    if prefix == "reco" and match_mode != "truth_to_reco":
                        keys[f"{obj_name}_matches_r2t"] = True
                        keys[f"{obj_name}_matches_r2t_overlap"] = True
                    if prefix == "truth" and match_mode != "reco_to_truth":
                        keys[f"{obj_name}_matches_t2r"] = True
                        keys[f"{obj_name}_matches_t2r_overlap"] = True

        self.update_keys(keys)

        # Initialize one CSV writer per object type
        for key in self.obj_keys:
            self.initialize_writer(key)

        assert len(self.writers), "Must request to save something."

    def process(self, data):
        """Store the information from one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over the keys to store
        other_prefix = {"reco": "truth", "truth": "reco"}
        for key in self.obj_keys:
            # Dispatch
            prefix, obj_type = key.split("_")
            other = other_prefix[prefix]
            attrs = self.attrs[key]
            lengths = self.lengths
            if self.match_mode is None or self.match_mode == f"{other}_to_{prefix}":
                # If there is no matches, save objects by themselves
                for i, obj in enumerate(data[key]):
                    self.append(key, **obj.scalar_dict(attrs, lengths))

            else:
                # If there are matches, combine the objects with their best
                # match on a single row
                match_suffix = f"{prefix[0]}2{other[0]}"
                match_key = f"{obj_type[:-1]}_matches_{match_suffix}"
                attrs_other = self.attrs[f"{other}_{obj_type}"]
                lengths_other = self.lengths  # TODO
                for idx, (obj_i, obj_j) in enumerate(data[match_key]):
                    src_dict = obj_i.scalar_dict(attrs, lengths)
                    if obj_j is not None:
                        tgt_dict = obj_j.scalar_dict(attrs_other, lengths_other)
                    else:
                        default_obj = self.default_objs[f"{other}_{obj_type}"]
                        tgt_dict = default_obj.scalar_dict(attrs_other, lengths_other)

                    src_dict = {f"{prefix}_{k}": v for k, v in src_dict.items()}
                    tgt_dict = {f"{other}_{k}": v for k, v in tgt_dict.items()}
                    overlap = data[f"{match_key}_overlap"][idx]

                    row_dict = {**src_dict, **tgt_dict}
                    row_dict.update({"match_overlap": overlap})
                    self.append(key, **row_dict)
